"""Creates histograms from ntuples with uproot."""

import logging
import pathlib
from typing import List, Optional

import boost_histogram as bh
import numpy as np
import uproot


log = logging.getLogger(__name__)


def with_uproot(
    ntuple_paths: List[pathlib.Path],
    pos_in_file: str,
    variable: str,
    bins: np.ndarray,
    *,
    weight: Optional[str] = None,
    selection_filter: Optional[str] = None,
) -> bh.Histogram:
    """Reads an ntuple with uproot, fills and returns a histogram with the observable.

    The paths may contain wildcards.

    Args:
        ntuple_paths (List[pathlib.Path]): list of paths to ntuples
        pos_in_file (str): name of tree within ntuple
        variable (str): variable to bin histogram in
        bins (np.ndarray): bin edges for histogram
        weight (Optional[str], optional): event weight to extract, defaults to None (no
            weights applied)
        selection_filter (Optional[str], optional): filter to be applied on events,
            defaults to None (no filter)

    Returns:
        bh.Histogram: histogram containing data
    """
    # concatenate the path to file and location within file with ":"
    paths_with_trees = [str(path) + ":" + pos_in_file for path in ntuple_paths]

    # determine whether the weight is a float or an expression
    # (for which a branch needs to be read)
    if weight is not None:
        try:
            float(weight)
            weight_is_expression = False
        except ValueError:
            # weight is not a float, need to evaluate the expression
            weight_is_expression = True
    else:
        # no weight specified, all weights are 1.0
        weight_is_expression = False
        weight = "1.0"

    log.debug(f"reading input ntuples {paths_with_trees}")

    if weight_is_expression:
        # need to read observables and weights
        array_generator = uproot.iterate(
            paths_with_trees, expressions=[variable, weight], cut=selection_filter
        )
        obs_list = []
        weight_list = []
        for arr in array_generator:
            obs_list.append(arr[variable].to_numpy())
            weight_list.append(arr[weight].to_numpy())
        observables = np.concatenate(obs_list)
        weights = np.concatenate(weight_list)

    else:
        # only need to read the observables
        array_generator = uproot.iterate(
            paths_with_trees, expressions=[variable], cut=selection_filter
        )
        obs_list = []
        for arr in array_generator:
            obs_list.append(arr[variable].to_numpy())
        observables = np.concatenate(obs_list)
        weights = np.ones_like(observables) * float(weight)

    # create and return histogram
    hist = bh.Histogram(bh.axis.Variable(bins), storage=bh.storage.Weight())
    hist.fill(observables, weight=weights)
    return hist



import copy
import pathlib
from unittest import mock

import pytest

from cabinetry import histo
from cabinetry import workspace


def test_WorkspaceBuilder():
    config = {"General": {"HistogramFolder": "path/"}}
    ws_builder = workspace.WorkspaceBuilder(config)
    assert ws_builder.config == config
    assert ws_builder.histogram_folder == pathlib.Path("path/")


def test_WorkspaceBuilder__data_sample():
    mc_sample = {"Name": "MC"}
    data_sample = {"Name": "Data", "Data": True}

    example_config = {
        "General": {"HistogramFolder": "path"},
        "Samples": [mc_sample, data_sample],
    }
    ws_builder = workspace.WorkspaceBuilder(example_config)
    assert ws_builder._data_sample() == data_sample

    config_two_data_samples = {
        "General": {"HistogramFolder": "path"},
        "Samples": [data_sample, data_sample],
    }
    ws_builder = workspace.WorkspaceBuilder(config_two_data_samples)
    with pytest.raises(ValueError, match="did not find exactly one data sample"):
        ws_builder._data_sample()


def test_WorkspaceBuilder__constant_parameter_setting():
    config_no_fixed = {"General": {"HistogramFolder": "path"}}
    ws_builder = workspace.WorkspaceBuilder(config_no_fixed)
    assert ws_builder._constant_parameter_setting("par") is None

    config_others_fixed = {
        "General": {
            "HistogramFolder": "path",
            "Fixed": [{"Name": "par_a", "Value": 1.2}],
        }
    }
    ws_builder = workspace.WorkspaceBuilder(config_others_fixed)
    assert ws_builder._constant_parameter_setting("par_b") is None

    config_par_fixed = {
        "General": {
            "HistogramFolder": "path",
            "Fixed": [{"Name": "par_a", "Value": 1.2}],
        }
    }
    ws_builder = workspace.WorkspaceBuilder(config_par_fixed)
    assert ws_builder._constant_parameter_setting("par_a") == 1.2


def test_WorkspaceBuilder_normfactor_modifiers():
    # could mock region_contains_modifier / sample_contains_modifier
    # one NF affects sample
    example_config = {
        "General": {"HistogramFolder": "path"},
        "NormFactors": [{"Name": "mu", "Samples": ["ABC", "DEF"]}],
    }
    region = {"Name": "SR"}
    sample = {"Name": "DEF"}
    expected_modifier = [{"data": None, "name": "mu", "type": "normfactor"}]
    ws_builder = workspace.WorkspaceBuilder(example_config)
    assert ws_builder.normfactor_modifiers(region, sample) == expected_modifier

    # no NF affects sample
    sample = {"Name": "GHI"}
    assert ws_builder.normfactor_modifiers(region, sample) == []

    # NF enters in region
    example_config = {
        "General": {"HistogramFolder": "path"},
        "NormFactors": [{"Name": "mu", "Regions": "SR"}],
    }
    ws_builder = workspace.WorkspaceBuilder(example_config)
    assert ws_builder.normfactor_modifiers(region, sample) == expected_modifier

    # no NF due to region
    region = {"Name": "CR"}
    assert ws_builder.normfactor_modifiers(region, sample) == []

    # multiple NFs affect sample
    example_config = {
        "General": {"HistogramFolder": "path"},
        "NormFactors": [{"Name": "mu"}, {"Name": "k"}],
    }
    sample = {"Name": "DEF"}
    expected_modifier = [
        {"data": None, "name": "mu", "type": "normfactor"},
        {"data": None, "name": "k", "type": "normfactor"},
    ]
    ws_builder = workspace.WorkspaceBuilder(example_config)
    assert ws_builder.normfactor_modifiers(region, sample) == expected_modifier


def test_WorkspaceBuilder_normalization_modifier():
    systematic = {
        "Name": "sys",
        "Up": {"Normalization": 0.1},
        "Down": {"Normalization": -0.05},
    }
    expected_modifier = {
        "name": "sys",
        "type": "normsys",
        "data": {"hi": 1.1, "lo": 0.95},
    }
    ws_builder = workspace.WorkspaceBuilder({"General": {"HistogramFolder": "path"}})
    assert ws_builder.normalization_modifier(systematic) == expected_modifier

    # ModifierName set
    systematic = {
        "Name": "sys",
        "Up": {"Normalization": 0.1},
        "Down": {"Normalization": -0.05},
        "ModifierName": "mod_name",
    }
    assert ws_builder.normalization_modifier(systematic)["name"] == "mod_name"


@mock.patch(
    "cabinetry.workspace.histo.Histogram.from_config",
    side_effect=[
        # without symmetrization: up, nominal, down
        histo.Histogram.from_arrays([0, 1, 2], [26.0, 24.0], [0.1, 0.1]),
        histo.Histogram.from_arrays([0, 1, 2], [20.0, 20.0], [0.1, 0.1]),
        histo.Histogram.from_arrays([0, 1, 2], [8.0, 12.0], [0.1, 0.1]),
        # for test of symmetrization: up and nominal
        histo.Histogram.from_arrays([0, 1, 2], [26.0, 24.0], [0.1, 0.1]),
        histo.Histogram.from_arrays([0, 1, 2], [20.0, 20.0], [0.1, 0.1]),
        # single bin for test of histosys being skipped (up and nominal)
        histo.Histogram.from_arrays([0, 1], [26.0], [0.1]),
        histo.Histogram.from_arrays([0, 1], [20.0], [0.1]),
    ],
)
def test_WorkspaceBuilder_normplusshape_modifiers(mock_histogram):
    # could mock Histogram.normalize_to_yield
    # up: 26, 24 (1.25*nom)
    # nominal: 20, 20
    # down: 8, 12 (0.5*nom)
    example_config = {"General": {"HistogramFolder": "path"}}
    ws_builder = workspace.WorkspaceBuilder(example_config)
    region = {"Name": "SR"}
    sample = {"Name": "Signal"}
    systematic = {"Name": "sys", "Up": {}, "Down": {}}
    # no symmetrization
    modifiers = ws_builder.normplusshape_modifiers(region, sample, systematic)
    assert modifiers == [
        {"name": "sys", "type": "normsys", "data": {"hi": 1.25, "lo": 0.5}},
        {
            "name": "sys",
            "type": "histosys",
            "data": {"hi_data": [20.8, 19.2], "lo_data": [16.0, 24.0]},
        },
    ]
    assert mock_histogram.call_args_list == [
        (
            (pathlib.Path("path"), region, sample, systematic),
            {"template": "Up", "modified": True},
        ),
        ((pathlib.Path("path"), region, sample, {}), {"modified": True}),
        (
            (pathlib.Path("path"), region, sample, systematic),
            {"template": "Down", "modified": True},
        ),
    ]

    # down template via symmetrized up template, ModifierName set
    systematic = {
        "Name": "sys",
        "Up": {},
        "Down": {"Symmetrize": True},
        "ModifierName": "mod_name",
    }
    modifiers = ws_builder.normplusshape_modifiers(region, sample, systematic)
    assert modifiers == [
        {"name": "mod_name", "type": "normsys", "data": {"hi": 1.25, "lo": 0.75}},
        {
            "name": "mod_name",
            "type": "histosys",
            "data": {"hi_data": [20.8, 19.2], "lo_data": [19.2, 20.8]},
        },
    ]
    assert mock_histogram.call_args_list[3:] == [
        (
            (pathlib.Path("path"), region, sample, systematic),
            {"template": "Up", "modified": True},
        ),
        ((pathlib.Path("path"), region, sample, {}), {"modified": True}),
    ]

    # single bin, causing histosys being skipped
    modifiers = ws_builder.normplusshape_modifiers(region, sample, systematic)
    assert modifiers == [
        {"name": "mod_name", "type": "normsys", "data": {"hi": 1.3, "lo": 0.7}},
    ]


@mock.patch(
    "cabinetry.workspace.WorkspaceBuilder.normplusshape_modifiers",
    return_value=[{"mock": "norm"}, {"mock": "shape"}],
)
@mock.patch(
    "cabinetry.workspace.WorkspaceBuilder.normalization_modifier",
    return_value={"mock": "normsys"},
)
def test_WorkspaceBuilder_sys_modifiers(mock_norm, mock_norm_shape):
    # could mock region_contains_modifier / sample_contains_modifier
    example_config = {
        "General": {"HistogramFolder": "path"},
        "Systematics": [
            {"Name": "norm", "Type": "Normalization"},
            {"Name": "norm_shape", "Type": "NormPlusShape"},
        ],
    }
    region = {"Name": "SR"}
    sample = {"Name": "Signal"}
    ws_builder = workspace.WorkspaceBuilder(example_config)
    assert ws_builder.sys_modifiers(region, sample) == [
        {"mock": "normsys"},
        {"mock": "norm"},
        {"mock": "shape"},
    ]
    assert mock_norm.call_args_list == [((example_config["Systematics"][0],), {})]
    assert mock_norm_shape.call_args_list == [
        ((region, sample, example_config["Systematics"][1]), {})
    ]

    # one systematic not present in region
    example_config_region_mismatch = copy.deepcopy(example_config)
    example_config_region_mismatch["Systematics"][0].update({"Regions": "CR"})
    ws_builder = workspace.WorkspaceBuilder(example_config_region_mismatch)
    assert ws_builder.sys_modifiers(region, sample) == [
        {"mock": "norm"},
        {"mock": "shape"},
    ]

    # one systematic not present in sample
    example_config_sample_mismatch = copy.deepcopy(example_config)
    example_config_sample_mismatch["Systematics"][1].update({"Samples": "Background"})
    ws_builder = workspace.WorkspaceBuilder(example_config_sample_mismatch)
    assert ws_builder.sys_modifiers(region, sample) == [{"mock": "normsys"}]

    # unsupported systematics type
    example_config_unsupported = {
        "General": {"HistogramFolder": "path"},
        "Systematics": [{"Name": "Normalization", "Type": "unknown"}],
    }
    ws_builder = workspace.WorkspaceBuilder(example_config_unsupported)
    with pytest.raises(
        NotImplementedError, match="not supporting other systematic types yet"
    ):
        ws_builder.sys_modifiers(region, sample)


@mock.patch(
    "cabinetry.workspace.histo.Histogram.from_config",
    return_value=histo.Histogram.from_arrays([0, 1, 2], [1.0, 2.0], [0.1, 0.1]),
)
@mock.patch(
    "cabinetry.configuration.region_contains_sample", side_effect=[True, False, True]
)
def test_WorkspaceBuilder_channels(mock_contains, mock_histogram):
    # should mock normfactor_modifiers / sys_modifiers
    example_config = {
        "General": {"HistogramFolder": "path"},
        "Regions": [{"Name": "region_1"}],
        "Samples": [{"Name": "signal"}, {"Data": True}],
        "NormFactors": [],
    }

    ws_builder = workspace.WorkspaceBuilder(example_config)
    channels = ws_builder.channels()
    expected_channels = [
        {
            "name": "region_1",
            "samples": [
                {
                    "name": "signal",
                    "data": [1.0, 2.0],
                    "modifiers": [
                        {
                            "name": "staterror_region_1",
                            "type": "staterror",
                            "data": [0.1, 0.1],
                        }
                    ],
                }
            ],
        }
    ]
    assert channels == expected_channels
    assert mock_contains.call_args_list == [
        ((example_config["Regions"][0], example_config["Samples"][0]), {})
    ]
    assert mock_histogram.call_args_list == [
        (
            (
                pathlib.Path("path"),
                example_config["Regions"][0],
                example_config["Samples"][0],
                {},
            ),
            {"modified": True},
        )
    ]

    # run again, this time region will not contain sample due to side_effect
    channels = ws_builder.channels()
    expected_channels = [{"name": "region_1", "samples": []}]
    assert channels == expected_channels
    assert mock_contains.call_count == 2
    assert mock_contains.call_args == (
        (example_config["Regions"][0], example_config["Samples"][0]),
        {},
    )
    # no calls to read histogram content
    assert mock_histogram.call_count == 1

    # staterror creation disabled
    example_config = {
        "General": {"HistogramFolder": "path"},
        "Regions": [{"Name": "region_1"}],
        "Samples": [{"Name": "signal", "DisableStaterror": True}],
        "NormFactors": [],
    }
    ws_builder = workspace.WorkspaceBuilder(example_config)
    channels = ws_builder.channels()
    assert channels[0]["samples"][0]["modifiers"] == []


def test_WorkspaceBuilder_measurements():
    example_config = {
        "General": {"Measurement": "fit", "HistogramFolder": "path"},
        "NormFactors": [
            {"Name": "NF", "Nominal": 1.0, "Bounds": [0.0, 5.0], "Fixed": False}
        ],
    }
    expected_measurement = [
        {
            "name": "fit",
            "config": {
                "poi": "",
                "parameters": [{"name": "NF", "inits": [1.0], "bounds": [[0.0, 5.0]]}],
            },
        }
    ]
    ws_builder = workspace.WorkspaceBuilder(example_config)
    assert ws_builder.measurements() == expected_measurement

    # no norm factor settings, POI specified
    example_config_no_NF_settings = {
        "General": {"Measurement": "fit", "POI": "mu", "HistogramFolder": "path"},
        "NormFactors": [{"Name": "NF"}],
    }
    expected_measurement_no_NF_settings = [
        {"name": "fit", "config": {"poi": "mu", "parameters": [{"name": "NF"}]}}
    ]
    ws_builder = workspace.WorkspaceBuilder(example_config_no_NF_settings)
    assert ws_builder.measurements() == expected_measurement_no_NF_settings

    # constant normfactor
    with mock.patch(
        "cabinetry.workspace.WorkspaceBuilder._constant_parameter_setting",
        return_value=1.2,
    ) as mock_find_const:
        expected_measurement_const_NF = [
            {
                "name": "fit",
                "config": {
                    "poi": "mu",
                    "parameters": [{"name": "NF", "fixed": True, "inits": [1.2]}],
                },
            }
        ]
        # same config, but patched function to treat NF as fixed
        ws_builder = workspace.WorkspaceBuilder(example_config_no_NF_settings)
        assert ws_builder.measurements() == expected_measurement_const_NF
        assert mock_find_const.call_args_list == [(("NF",), {})]

    # constant systematic
    with mock.patch(
        "cabinetry.workspace.WorkspaceBuilder._constant_parameter_setting",
        return_value=1.2,
    ) as mock_find_const:
        example_config_const_sys = {
            "General": {
                "Measurement": "fit",
                "Fixed": [{"Name": "par_A", "Value": 1.2}],
                "HistogramFolder": "path",
            },
            "Systematics": [{"Name": "par_A"}],
        }
        expected_measurement_const_sys = [
            {
                "name": "fit",
                "config": {
                    "poi": "",
                    "parameters": [{"name": "par_A", "fixed": True, "inits": [1.2]}],
                },
            }
        ]
        ws_builder = workspace.WorkspaceBuilder(example_config_const_sys)
        assert ws_builder.measurements() == expected_measurement_const_sys
        assert mock_find_const.call_args_list == [(("par_A",), {})]

    # no constant systematic
    with mock.patch(
        "cabinetry.workspace.WorkspaceBuilder._constant_parameter_setting",
        return_value=None,
    ) as mock_find_const:
        example_config_sys = {
            "General": {"Measurement": "fit", "HistogramFolder": "path"},
            "Systematics": [{"Name": "par_A"}],
        }
        expected_measurement_sys = [
            {"name": "fit", "config": {"poi": "", "parameters": []}}
        ]
        ws_builder = workspace.WorkspaceBuilder(example_config_sys)
        assert ws_builder.measurements() == expected_measurement_sys
        assert mock_find_const.call_args_list == [(("par_A",), {})]


@mock.patch(
    "cabinetry.workspace.histo.Histogram.from_config",
    side_effect=[
        histo.Histogram.from_arrays([0, 1, 2], [1.0, 2.0], [0.1, 0.1]),
        histo.Histogram.from_arrays([0, 1], [5.0], [0.1]),
        histo.Histogram.from_arrays([0, 1], [3.0], [0.1]),
    ],
)
def test_WorkspaceBuilder_observations(mock_histogram):
    # could mock _data_sample
    # create observations list from config
    example_config = {
        "General": {"HistogramFolder": "path"},
        "Samples": [{"Name": "data", "Data": True}],
        "Regions": [{"Name": "test_region"}],
    }
    ws_builder = workspace.WorkspaceBuilder(example_config)
    obs = ws_builder.observations()
    expected_obs = [{"name": "test_region", "data": [1.0, 2.0]}]
    assert obs == expected_obs
    assert mock_histogram.call_args_list == [
        (
            (
                pathlib.Path("path"),
                example_config["Regions"][0],
                example_config["Samples"][0],
                {},
            ),
            {"modified": True},
        )
    ]

    # multiple channels
    multi_channel_config = {
        "General": {"HistogramFolder": "path"},
        "Samples": [{"Name": "data", "Data": True}],
        "Regions": [{"Name": "test_region"}, {"Name": "other_region"}],
    }
    ws_builder = workspace.WorkspaceBuilder(multi_channel_config)
    obs = ws_builder.observations()
    expected_obs = [
        {"name": "test_region", "data": [5.0]},
        {"name": "other_region", "data": [3.0]},
    ]
    assert obs == expected_obs
    assert mock_histogram.call_args_list[1:] == [
        (
            (
                pathlib.Path("path"),
                multi_channel_config["Regions"][0],
                multi_channel_config["Samples"][0],
                {},
            ),
            {"modified": True},
        ),
        (
            (
                pathlib.Path("path"),
                multi_channel_config["Regions"][1],
                multi_channel_config["Samples"][0],
                {},
            ),
            {"modified": True},
        ),
    ]


@mock.patch(
    "cabinetry.workspace.WorkspaceBuilder.observations",
    return_value=[{"name: observations"}],
)
@mock.patch(
    "cabinetry.workspace.WorkspaceBuilder.measurements",
    return_value=[{"name: measurement"}],
)
@mock.patch(
    "cabinetry.workspace.WorkspaceBuilder.channels", return_value=[{"name: channel"}]
)
def test_WorkspaceBuilder_build(mock_channels, mock_measuremets, mock_observations):
    ws_builder = workspace.WorkspaceBuilder({"General": {"HistogramFolder": "path"}})
    ws = ws_builder.build()
    ws_expected = {
        "channels": [{"name: channel"}],
        "measurements": [{"name: measurement"}],
        "observations": [{"name: observations"}],
        "version": "1.0.0",
    }
    assert ws == ws_expected


def test_build():
    minimal_config = {"General": {"Measurement": "test"}}

    mock_builder_instance = mock.MagicMock()
    mock_builder_instance.build.return_value = {"name": "workspace"}

    with mock.patch(
        "cabinetry.workspace.WorkspaceBuilder", return_value=mock_builder_instance
    ) as mock_builder:
        # without validation
        ws = workspace.build(minimal_config, with_validation=False)
        ws_expected = {"name": "workspace"}
        assert ws == ws_expected
        assert mock_builder.call_args_list == [((minimal_config,), {})]

        # including validation
        with mock.patch("cabinetry.workspace.validate") as mock_validate:
            ws = workspace.build(minimal_config, with_validation=True)
            assert ws == ws_expected
            assert mock_validate.call_args_list == [((ws,), {})]


@mock.patch("cabinetry.workspace.pyhf.Workspace")
def test_validate(mock_validate):
    test_ws = {"workspace": "test"}
    workspace.validate(test_ws)
    assert mock_validate.call_args_list == [((test_ws,), {})]


def test_save(tmp_path):
    # save to subfolder that needs to be created
    fname = tmp_path / "subdir" / "ws.json"
    ws = {"version": "1.0.0"}
    workspace.save(ws, fname)
    assert workspace.load(fname) == ws


def test_load(tmp_path):
    fname = tmp_path / "ws.json"
    ws = {"version": "1.0.0"}
    workspace.save(ws, fname)
    assert workspace.load(fname) == ws



import numpy as np

from cabinetry.contrib import histogram_creator


def test_with_uproot(tmp_path, utils):
    fnames = [tmp_path / "test.root"]
    treename = "tree"
    varname = "var"
    var_array = np.asarray([1.1, 2.3, 3.0, 3.2])
    weightname_write = "weight"
    weight_array = np.asarray([1.0, 1.0, 2.0, 1.0])
    bins = np.asarray([1, 2, 3, 4])
    # create something to read
    utils.create_ntuple(
        fnames[0], treename, varname, var_array, weightname_write, weight_array
    )

    # read - no weights or selection
    hist = histogram_creator.with_uproot(fnames, treename, varname, bins)
    assert np.allclose(hist.values(), [1, 1, 2])
    assert np.allclose(np.sqrt(hist.variances()), [1, 1, 1.41421356])

    # read - with selection cut
    selection = "var < 3.1"
    hist = histogram_creator.with_uproot(
        fnames, treename, varname, bins, selection_filter=selection
    )
    assert np.allclose(hist.values(), [1, 1, 1])
    assert np.allclose(np.sqrt(hist.variances()), [1, 1, 1])

    # read - with weight
    weightname_apply = "weight*2"
    hist = histogram_creator.with_uproot(
        fnames, treename, varname, bins, weight=weightname_apply
    )
    assert np.allclose(hist.values(), [2, 2, 6])
    assert np.allclose(np.sqrt(hist.variances()), [2, 2, 4.47213595])

    # read - with weight that is only a float
    weight_float = "2.0"
    hist = histogram_creator.with_uproot(
        fnames, treename, varname, bins, weight=weight_float
    )
    assert np.allclose(hist.values(), [2, 2, 4])
    assert np.allclose(np.sqrt(hist.variances()), [2, 2, 2.82842712])

    # create an additional file to test wildcards / input lists
    extra_path = tmp_path / "test_2.root"
    utils.create_ntuple(
        extra_path, treename, varname, var_array, weightname_write, weight_array
    )

    # read - with wildcard, matching two files
    fnames = [tmp_path / "test*.root"]
    hist = histogram_creator.with_uproot(fnames, treename, varname, bins)
    assert np.allclose(hist.values(), [2, 2, 4])
    assert np.allclose(np.sqrt(hist.variances()), [1.41421356, 1.41421356, 2])

    # read - with two list entries
    fnames = [tmp_path / "test.root", tmp_path / "test_2.root"]
    hist = histogram_creator.with_uproot(fnames, treename, varname, bins)
    assert np.allclose(hist.values(), [2, 2, 4])
    assert np.allclose(np.sqrt(hist.variances()), [1.41421356, 1.41421356, 2])



import logging
import pathlib
from unittest import mock

from cabinetry import templates


@mock.patch("cabinetry.route.apply_to_all_templates")
@mock.patch("cabinetry.templates.builder._Builder")
def test_build(mock_builder, mock_apply):
    config = {"General": {"HistogramFolder": "path/", "InputPath": "file.root"}}
    method = "uproot"

    # no router
    templates.build(config, method=method)
    assert mock_builder.call_args_list == [
        ((pathlib.Path("path/"), "file.root", method), {})
    ]
    assert mock_apply.call_count == 1
    config_call, func_call = mock_apply.call_args[0]
    assert config_call == config
    assert func_call._extract_mock_name() == "_Builder()._create_histogram"
    assert mock_apply.call_args[1] == {"match_func": None}

    # including a router
    mock_router = mock.MagicMock()
    templates.build(config, method=method, router=mock_router)

    # verify wrapper was set
    assert (
        mock_router.template_builder_wrapper._extract_mock_name()
        == "_Builder()._wrap_custom_template_builder"
    )

    assert mock_apply.call_count == 2  # 1 from before
    config_call, func_call = mock_apply.call_args[0]
    assert config_call == config
    assert func_call._extract_mock_name() == "_Builder()._create_histogram"
    assert mock_apply.call_args[1] == {
        "match_func": mock_router._find_template_builder_match
    }


@mock.patch("cabinetry.route.apply_to_all_templates")
@mock.patch("cabinetry.templates.collector._collector", return_value="func")
def test_collect(mock_collector, mock_apply, caplog):
    caplog.set_level(logging.DEBUG)
    config = {
        "General": {
            "HistogramFolder": "path/",
            "InputPath": "f.root:{VariationPath}",
            "VariationPath": "nominal",
        }
    }
    method = "uproot"

    templates.collect(config, method=method)
    assert mock_collector.call_args_list == [
        ((pathlib.Path("path/"), "f.root:{VariationPath}", "nominal", method), {})
    ]
    assert mock_apply.call_args_list == [((config, "func"), {})]

    caplog.clear()

    # no VariationPath in general settings
    config = {
        "General": {"HistogramFolder": "path/", "InputPath": "f.root:{VariationPath}"}
    }
    templates.collect(config, method=method)
    assert 'no VariationPath specified in general settings, defaulting to ""' in [
        rec.message for rec in caplog.records
    ]
    assert mock_collector.call_args == (
        (pathlib.Path("path/"), "f.root:{VariationPath}", "", method),
        {},
    )
    caplog.set_level(logging.DEBUG)


@mock.patch("cabinetry.route.apply_to_all_templates")
@mock.patch("cabinetry.templates.postprocessor._postprocessor", return_value="func")
def test_run(mock_postprocessor, mock_apply):
    config = {"General": {"HistogramFolder": "path/"}}

    templates.postprocess(config)
    assert mock_postprocessor.call_args_list == [((pathlib.Path("path/"),), {})]
    assert mock_apply.call_args_list == [((config, "func"), {})]



import logging

import boost_histogram as bh
import numpy as np
import pytest

from cabinetry import histo


class ExampleHistograms:
    """a collection of different kinds of histograms"""

    @staticmethod
    def normal():
        bins = [1, 2, 3]
        yields = [1.0, 2.0]
        stdev = [0.1, 0.2]
        return bins, yields, stdev

    @staticmethod
    def single_bin():
        bins = [1, 2]
        yields = [1.0]
        stdev = [0.1]
        return bins, yields, stdev

    @staticmethod
    def empty_bin():
        bins = [1, 2, 3]
        yields = [1.0, 0.0]
        stdev = [0.1, 0.2]
        return bins, yields, stdev

    @staticmethod
    def nan_stdev_empty_bin():
        bins = [1, 2, 3]
        yields = [1.0, 0.0]
        stdev = [0.1, float("NaN")]
        return bins, yields, stdev

    @staticmethod
    def nan_stdev_nonempty_bin():
        bins = [1, 2, 3]
        yields = [1.0, 2.0]
        stdev = [0.1, float("NaN")]
        return bins, yields, stdev

    @staticmethod
    def wrong_bin_number():
        bins = [1, 2, 3]
        yields = [1.0]
        stdev = [0.1]
        return bins, yields, stdev

    @staticmethod
    def yields_stdev_mismatch():
        bins = [1, 2, 3]
        yields = [1.0, 2.0]
        stdev = [0.1]
        return bins, yields, stdev


@pytest.fixture
def example_histograms():
    return ExampleHistograms


class HistogramHelpers:
    @staticmethod
    def assert_equal(h1, h2):
        assert np.allclose(h1.yields, h2.yields)
        assert np.allclose(h1.stdev, h2.stdev)
        assert np.allclose(h1.bins, h2.bins)


@pytest.fixture
def histogram_helpers():
    return HistogramHelpers


# test properties
def test_Histogram(example_histograms):
    bins, yields, stdev = example_histograms.normal()
    h_orig = bh.Histogram(bh.axis.Variable(bins), storage=bh.storage.Weight())
    h_orig[...] = np.stack([np.asarray(yields), np.asarray(stdev) ** 2], axis=-1)
    h = histo.Histogram(h_orig)
    np.testing.assert_equal(h.bins, bins)
    np.testing.assert_equal(h.yields, yields)
    np.testing.assert_equal(h.stdev, stdev)
    new_yields = np.asarray([3, 4])
    new_stdev = np.asarray([0.5, 0.5])
    h.yields = new_yields
    h.stdev = new_stdev
    np.testing.assert_equal(h.yields, new_yields)
    np.testing.assert_equal(h.stdev, new_stdev)


def test_Histogram_from_arrays(example_histograms):
    bins, yields, stdev = example_histograms.normal()
    h = histo.Histogram.from_arrays(bins, yields, stdev)
    assert np.allclose(h.yields, yields)
    assert np.allclose(h.stdev, stdev)
    assert np.allclose(h.bins, bins)

    with pytest.raises(ValueError, match="bin edges need one more entry than yields"):
        histo.Histogram.from_arrays(*example_histograms.wrong_bin_number())

    with pytest.raises(
        ValueError, match="yields and stdev need to have the same shape"
    ):
        histo.Histogram.from_arrays(*example_histograms.yields_stdev_mismatch())


def test_Histogram_from_path(tmp_path, caplog, example_histograms, histogram_helpers):
    caplog.set_level(logging.INFO)
    h_ref = histo.Histogram.from_arrays(*example_histograms.normal())
    h_ref.save(tmp_path)

    # load the original histogram
    h_from_path = histo.Histogram.from_path(tmp_path, modified=False)
    histogram_helpers.assert_equal(h_ref, h_from_path)

    # try loading a modified one, without success since it does not exist
    h_from_path_modified = histo.Histogram.from_path(tmp_path, modified=True)
    expected_info = (
        f"no modified histogram {str(tmp_path)}_modified.npz found, "
        "loading un-modified histogram"
    )
    assert expected_info in [rec.message for rec in caplog.records]
    caplog.clear()

    # successfully load a modified histogram
    histo_name = "histo"
    fname_modified = tmp_path / (histo_name + "_modified")
    fname_original = tmp_path / histo_name
    h_ref.save(fname_modified)
    # load the modified histogram by specifying the original name, which should produce
    # no warning
    h_from_path_modified = histo.Histogram.from_path(fname_original, modified=True)
    histogram_helpers.assert_equal(h_ref, h_from_path_modified)
    assert [rec.message for rec in caplog.records] == []
    caplog.clear()


def test_Histogram_from_config(tmp_path, example_histograms, histogram_helpers):
    # could mock histo.name here
    h_ref = histo.Histogram.from_arrays(*example_histograms.normal())
    histo_path = tmp_path / "region_sample.npz"
    h_ref.save(histo_path)

    region = {"Name": "region"}
    sample = {"Name": "sample"}
    systematic = {}
    h_from_path = histo.Histogram.from_config(
        tmp_path, region, sample, systematic, modified=False
    )
    histogram_helpers.assert_equal(h_ref, h_from_path)

    # non-nominal and modified template
    histo_path = tmp_path / "region_sample_sys_Up_modified.npz"
    h_ref.save(histo_path)

    systematic = {"Name": "sys"}
    h_from_path = histo.Histogram.from_config(
        tmp_path, region, sample, systematic, template="Up"
    )
    histogram_helpers.assert_equal(h_ref, h_from_path)


def test_Histogram_save(tmp_path, example_histograms, histogram_helpers):
    hist = histo.Histogram.from_arrays(*example_histograms.normal())
    hist.save(tmp_path)
    hist_loaded = histo.Histogram.from_path(tmp_path, modified=False)
    histogram_helpers.assert_equal(hist, hist_loaded)
    # add a subdirectory that needs to be created for histogram saving
    fname = tmp_path / "subdir" / "file"
    hist.save(fname)
    hist_loaded = histo.Histogram.from_path(fname, modified=False)
    histogram_helpers.assert_equal(hist, hist_loaded)


def test_Histogram_validate(caplog, example_histograms):
    caplog.set_level(logging.DEBUG)
    name = "test_histo"

    # should cause no warnings
    histo.Histogram.from_arrays(*example_histograms.normal()).validate(name)
    histo.Histogram.from_arrays(*example_histograms.single_bin()).validate(name)
    assert [rec.message for rec in caplog.records] == []
    caplog.clear()

    # check for empty bin warning
    histo.Histogram.from_arrays(*example_histograms.empty_bin()).validate(name)
    assert "test_histo has empty bins: [1]" in [rec.message for rec in caplog.records]
    caplog.clear()

    # check for ill-defined stat uncertainty warning
    histo.Histogram.from_arrays(*example_histograms.nan_stdev_empty_bin()).validate(
        name
    )
    assert "test_histo has empty bins: [1]" in [rec.message for rec in caplog.records]
    assert "test_histo has bins with ill-defined stat. unc.: [1]" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # check for ill-defined stat uncertainty warning with non-zero bin
    histo.Histogram.from_arrays(*example_histograms.nan_stdev_nonempty_bin()).validate(
        name
    )
    assert "test_histo has bins with ill-defined stat. unc.: [1]" in [
        rec.message for rec in caplog.records
    ]
    assert "test_histo has non-empty bins with ill-defined stat. unc.: [1]" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()


def test_Histogram_normalize_to_yield(example_histograms):
    # requires boost_histogram >= 1.0.2 for correct stdev scaling, but this feature is
    # not currently needed in cabinetry
    hist = histo.Histogram.from_arrays(*example_histograms.empty_bin())
    var_hist = histo.Histogram.from_arrays(*example_histograms.normal())
    factor = var_hist.normalize_to_yield(hist)
    assert factor == 3.0
    np.testing.assert_equal(var_hist.yields, np.asarray([1 / 3, 2 / 3]))
    assert np.allclose(var_hist.stdev, np.asarray([0.1 / 3, 0.2 / 3]))


def test_name():
    region = {"Name": "Region"}
    sample = {"Name": "Sample"}
    systematic = {"Name": "Systematic"}
    template = "Up"
    assert (
        histo.name(region, sample, systematic, template=template)
        == "Region_Sample_Systematic_Up"
    )

    # no template specified
    region = {"Name": "Region 1"}
    sample = {"Name": "Sample 1"}
    systematic = {"Name": "Systematic 1"}
    assert histo.name(region, sample, systematic) == "Region-1_Sample-1"

    # nominal template requested explicitly
    assert histo.name(region, sample, systematic, template=None) == "Region-1_Sample-1"



import pathlib
from unittest import mock

import matplotlib.pyplot as plt

from cabinetry.visualize import utils


def test__save_and_close(tmp_path):
    fig = plt.figure()
    fname = tmp_path / "fig.pdf"
    utils._save_and_close(fig, fname, False)
    assert fname.is_file()  # file was saved
    assert len(plt.get_fignums()) == 1  # figure is open

    fig = plt.figure()  # new figure to test closing
    utils._save_and_close(fig, fname, True)
    assert len(plt.get_fignums()) == 1  # previous figure still open
    plt.close("all")

    fig = mock.MagicMock()
    utils._save_and_close(fig, None, True)
    assert not fig.savefig.called  # no path provided, so figure was not saved


def test__log_figure_path():
    path = pathlib.Path("path/to_file.pdf")
    assert utils._log_figure_path(path) == pathlib.Path("path/to_file_log.pdf")

    # input is None
    path = None
    assert utils._log_figure_path(path) is None



import logging.config

import boost_histogram as bh
import numpy as np
import pytest
import uproot


def pytest_sessionstart():
    # suppress verbose DEBUG level output from matplotlib when tests fail
    logging.config.dictConfig(
        {
            "loggers": {"matplotlib": {"level": "INFO"}},
            "disable_existing_loggers": False,
            "version": 1,
        }
    )


class Utils:
    @staticmethod
    def create_ntuple(fname, treename, varname, var_array, weightname, weight_array):
        with uproot.recreate(fname) as f:
            f[treename] = {varname: var_array, weightname: weight_array}

    @staticmethod
    def create_histogram(fname, histname, bins, yields, stdev):
        hist = bh.Histogram(bh.axis.Variable(bins), storage=bh.storage.Weight())
        hist[...] = np.stack([yields, stdev**2], axis=-1)
        with uproot.recreate(fname) as f:
            f[histname] = hist


@pytest.fixture
def utils():
    return Utils


@pytest.fixture
def example_spec():
    spec = {
        "channels": [
            {
                "name": "Signal Region",
                "samples": [
                    {
                        "data": [51.8],
                        "modifiers": [
                            {
                                "data": [2.6],
                                "name": "staterror_Signal-Region",
                                "type": "staterror",
                            },
                            {
                                "data": None,
                                "name": "Signal strength",
                                "type": "normfactor",
                            },
                        ],
                        "name": "Signal",
                    }
                ],
            }
        ],
        "measurements": [
            {
                "config": {
                    "parameters": [
                        {
                            "name": "staterror_Signal-Region",
                            "fixed": True,
                            "inits": [1.1],
                        }
                    ],
                    "poi": "Signal strength",
                },
                "name": "My fit",
            }
        ],
        "observations": [{"data": [475], "name": "Signal Region"}],
        "version": "1.0.0",
    }
    return spec


@pytest.fixture
def example_spec_multibin():
    spec = {
        "channels": [
            {
                "name": "region_1",
                "samples": [
                    {
                        "data": [25, 5],
                        "modifiers": [
                            {
                                "data": [5, 2],
                                "name": "staterror_region_1",
                                "type": "staterror",
                            },
                            {
                                "data": None,
                                "name": "Signal strength",
                                "type": "normfactor",
                            },
                        ],
                        "name": "Signal",
                    }
                ],
            },
            {
                "name": "region_2",
                "samples": [
                    {
                        "data": [8],
                        "modifiers": [
                            {
                                "data": [1],
                                "name": "staterror_region_2",
                                "type": "staterror",
                            },
                            {
                                "data": None,
                                "name": "Signal strength",
                                "type": "normfactor",
                            },
                        ],
                        "name": "Signal",
                    }
                ],
            },
        ],
        "measurements": [
            {"config": {"parameters": [], "poi": "Signal strength"}, "name": "My fit"}
        ],
        "observations": [
            {"data": [35, 8], "name": "region_1"},
            {"data": [10], "name": "region_2"},
        ],
        "version": "1.0.0",
    }
    return spec


@pytest.fixture
def example_spec_shapefactor():
    spec = {
        "channels": [
            {
                "name": "Signal Region",
                "samples": [
                    {
                        "data": [20, 10],
                        "modifiers": [
                            {
                                "data": None,
                                "name": "shape factor",
                                "type": "shapefactor",
                            },
                            {
                                "data": None,
                                "name": "Signal strength",
                                "type": "normfactor",
                            },
                        ],
                        "name": "Signal",
                    }
                ],
            }
        ],
        "measurements": [
            {
                "config": {"parameters": [], "poi": "Signal strength"},
                "name": "shapefactor fit",
            }
        ],
        "observations": [{"data": [25, 8], "name": "Signal Region"}],
        "version": "1.0.0",
    }
    return spec


@pytest.fixture
def example_spec_with_background():
    spec = {
        "channels": [
            {
                "name": "Signal Region",
                "samples": [
                    {
                        "data": [50],
                        "modifiers": [
                            {
                                "data": [5],
                                "name": "staterror_Signal-Region",
                                "type": "staterror",
                            },
                            {
                                "data": None,
                                "name": "Signal strength",
                                "type": "normfactor",
                            },
                        ],
                        "name": "Signal",
                    },
                    {
                        "data": [150],
                        "modifiers": [
                            {
                                "data": [7],
                                "name": "staterror_Signal-Region",
                                "type": "staterror",
                            }
                        ],
                        "name": "Background",
                    },
                ],
            }
        ],
        "measurements": [
            {
                "config": {
                    "parameters": [
                        {"name": "Signal strength", "bounds": [[0, 10]], "inits": [1.0]}
                    ],
                    "poi": "Signal strength",
                },
                "name": "signal plus background",
            }
        ],
        "observations": [{"data": [160], "name": "Signal Region"}],
        "version": "1.0.0",
    }
    return spec


@pytest.fixture
def example_spec_no_aux():
    spec = {
        "channels": [
            {
                "name": "Signal Region",
                "samples": [
                    {
                        "data": [60],
                        "modifiers": [
                            {
                                "data": None,
                                "name": "Signal strength",
                                "type": "normfactor",
                            }
                        ],
                        "name": "Signal",
                    }
                ],
            }
        ],
        "measurements": [
            {
                "config": {"parameters": [], "poi": "Signal strength"},
                "name": "no auxdata",
            }
        ],
        "observations": [{"data": [65], "name": "Signal Region"}],
        "version": "1.0.0",
    }
    return spec


@pytest.fixture
def example_spec_lumi():
    spec = {
        "channels": [
            {
                "name": "Signal Region",
                "samples": [
                    {
                        "data": [35],
                        "modifiers": [
                            {
                                "data": None,
                                "name": "Signal strength",
                                "type": "normfactor",
                            },
                            {"data": None, "name": "lumi", "type": "lumi"},
                        ],
                        "name": "Signal",
                    }
                ],
            }
        ],
        "measurements": [
            {
                "config": {
                    "parameters": [
                        {
                            "auxdata": [1.0],
                            "inits": [1.05],
                            "name": "lumi",
                            "sigmas": [0.02],
                        }
                    ],
                    "poi": "Signal strength",
                },
                "name": "lumi modifier",
            }
        ],
        "observations": [{"data": [35], "name": "Signal Region"}],
        "version": "1.0.0",
    }
    return spec


@pytest.fixture
def example_spec_modifiers():
    spec = {
        "channels": [
            {
                "name": "SR",
                "samples": [
                    {
                        "data": [35.0, 30.0],
                        "modifiers": [
                            {"data": None, "name": "mu", "type": "normfactor"},
                            {
                                "data": None,
                                "name": "mu_shapefactor",
                                "type": "shapefactor",
                            },
                            {
                                "data": {"hi": 0.8, "lo": 1.2},
                                "name": "normsys",
                                "type": "normsys",
                            },
                            {
                                "data": {
                                    "hi_data": [36.0, 31.0],
                                    "lo_data": [34.0, 29.0],
                                },
                                "name": "histosys",
                                "type": "histosys",
                            },
                            {
                                "data": [1.0, 2.0],
                                "name": "staterror_SR",
                                "type": "staterror",
                            },
                            {
                                "data": [1.0, 2.0],
                                "name": "shapesys_SR",
                                "type": "shapesys",
                            },
                            {"data": None, "name": "lumi", "type": "lumi"},
                        ],
                        "name": "Signal",
                    }
                ],
            }
        ],
        "measurements": [
            {
                "config": {
                    "parameters": [
                        {
                            "auxdata": [1.0],
                            "inits": [1.0],
                            "name": "lumi",
                            "sigmas": [0.02],
                        }
                    ],
                    "poi": "mu",
                },
                "name": "lumi modifier",
            }
        ],
        "observations": [{"data": [35, 30], "name": "SR"}],
        "version": "1.0.0",
    }
    return spec


@pytest.fixture
def example_spec_zero_staterror():
    spec = {
        "channels": [
            {
                "name": "SR",
                "samples": [
                    {
                        "data": [5.0, 0.0],
                        "modifiers": [
                            {"data": None, "name": "mu", "type": "normfactor"},
                            {
                                "data": [1.0, 0.0],
                                "name": "staterror_SR",
                                "type": "staterror",
                            },
                        ],
                        "name": "Signal",
                    }
                ],
            }
        ],
        "measurements": [
            {
                "config": {
                    "parameters": [],
                    "poi": "mu",
                },
                "name": "zero staterror",
            }
        ],
        "observations": [{"data": [5, 0], "name": "SR"}],
        "version": "1.0.0",
    }
    return spec


# code below allows marking tests as slow and adds --runslow to run them
# implemented following https://docs.pytest.org/en/6.2.x/example/simple.html
def pytest_addoption(parser):
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)



"""Provides a histogram class based on boost-histogram."""

import logging
import os
import pathlib
from typing import Any, Dict, List, Literal, Optional, Type, TypeVar, Union

import boost_histogram as bh
import numpy as np

import cabinetry


log = logging.getLogger(__name__)


H = TypeVar("H", bound="Histogram")


class Histogram(bh.Histogram, family=cabinetry):
    """Holds histogram information, extends boost_histogram.Histogram."""

    @classmethod
    def from_arrays(
        cls: Type[H],
        bins: Union[List[float], np.ndarray],
        yields: Union[List[float], np.ndarray],
        stdev: Union[List[float], np.ndarray],
    ) -> H:
        """Constructs a histogram from arrays of yields and uncertainties.

        The input can be lists of ints or floats, or numpy.ndarrays.

        Args:
            bins (Union[List[float], np.ndarray]): edges of histogram bins
            yields (Union[List[float], np.ndarray]): yield per histogram bin
            stdev (Union[List[float], np.ndarray]): statistical uncertainty of yield per
                bin

        Raises:
            ValueError: when amount of bins specified via bin edges and bin contents do
                not match
            ValueError: when length of yields and stdev do not match

        Returns:
            cabinetry.histo.Histogram: the histogram instance
        """
        if len(bins) != len(yields) + 1:
            raise ValueError("bin edges need one more entry than yields")
        if len(yields) != len(stdev):
            raise ValueError("yields and stdev need to have the same shape")

        out = cls(
            bh.axis.Variable(bins, underflow=False, overflow=False),
            storage=bh.storage.Weight(),
        )
        yields = np.asarray(yields)
        stdev = np.asarray(stdev)
        out[...] = np.stack([yields, stdev**2], axis=-1)
        return out

    @classmethod
    def from_path(
        cls: Type[H], histo_path: pathlib.Path, *, modified: bool = True
    ) -> H:
        """Builds a histogram from disk.

        Loads the "modified" version of the histogram by default (which received post-
        processing).

        Args:
            histo_path (pathlib.Path): where the histogram is located
            modified (bool, optional): whether to load the modified histogram (after
                post-processing), defaults to True

        Returns:
            cabinetry.histo.Histogram: the loaded histogram
        """
        if modified:
            histo_path_modified = histo_path.parent / (histo_path.name + "_modified")
            if not histo_path_modified.with_suffix(".npz").exists():
                log.info(
                    f"no modified histogram {histo_path_modified.with_suffix('.npz')} "
                    "found, loading un-modified histogram"
                )
            else:
                histo_path = histo_path_modified
        histogram_npz = np.load(histo_path.with_suffix(".npz"))
        bins = histogram_npz["bins"]
        yields = histogram_npz["yields"]
        stdev = histogram_npz["stdev"]
        return cls.from_arrays(bins, yields, stdev)

    @classmethod
    def from_config(
        cls: Type[H],
        histo_folder: Union[str, pathlib.Path],
        region: Dict[str, Any],
        sample: Dict[str, Any],
        systematic: Dict[str, Any],
        *,
        template: Optional[Literal["Up", "Down"]] = None,
        modified: bool = True,
    ) -> H:
        """Loads a histogram, using information specified in the configuration file.

        To find the histogram, need to provide the folder the histogram is located in
        and the relevant information from the config: region, sample, systematic,
        template.

        Args:
            histo_folder (Union[str, patlib.Path]): folder containing all histograms
            region (Dict[str, Any]): containing all region information
            sample (Dict[str, Any]): containing all sample information
            systematic (Dict[str, Any]): containing all systematic information
            template (Optional[Literal["Up", "Down"]], optional): which template to
                consider: "Up", "Down", None for the nominal case, defaults to None
            modified (bool, optional): whether to load the modified histogram (after
                post-processing), defaults to True

        Returns:
            cabinetry.histo.Histogram: the loaded histogram
        """
        # find the histogram name given config information, and then load the histogram
        histo_name = name(region, sample, systematic, template=template)
        histo_path = pathlib.Path(histo_folder) / histo_name
        return cls.from_path(histo_path, modified=modified)

    @property
    def yields(self) -> np.ndarray:
        """Returns the yields per histogram bin.

        Returns:
            np.ndarray: yields per bin
        """
        return self.values()

    @yields.setter
    def yields(self, value: np.ndarray) -> None:
        """Updates the yields per bin.

        Args:
            value (np.ndarray): yields to set
        """
        self.view().value = value  # type: ignore[union-attr]

    @property
    def stdev(self) -> np.ndarray:
        """Returns the stat. uncertainty per histogram bin.

        Returns:
            np.ndarray: stat. uncertainty per bin
        """
        return np.sqrt(self.variances())  # type: ignore[arg-type]

    @stdev.setter
    def stdev(self, value: np.ndarray) -> None:
        """Updates the variance (by specifying the standard deviation).

        Args:
            value (np.ndarray): the standard deviation
        """
        self.view().variance = value**2  # type: ignore[misc,union-attr]

    @property
    def bins(self) -> np.ndarray:
        """Returns the bin edges.

        Returns:
            np.ndarray: bin edges
        """
        return self.axes[0].edges

    def save(self, histo_path: pathlib.Path) -> None:
        """Saves a histogram to disk.

        Args:
            histo_path (pathlib.Path): where to save the histogram
        """
        log.debug(f"saving histogram to {histo_path.with_suffix('.npz')}")

        # create output directory if it does not exist yet
        if not os.path.exists(histo_path.parent):
            os.mkdir(histo_path.parent)
        np.savez(
            histo_path.with_suffix(".npz"),
            yields=self.yields,
            stdev=self.stdev,
            bins=self.bins,
        )

    def validate(self, name: str) -> None:
        """Runs consistency checks on a histogram.

        Checks for empty bins and ill-defined statistical uncertainties. Logs warnings
        if issues are founds, but does not raise exceptions.

        Args:
            name (str): name of the histogram for logging purposes
        """
        # check for empty bins
        # using np.atleast_1d to fix deprecation warning, even though the
        # input should never need it
        empty_bins = np.where(np.atleast_1d(self.yields) == 0.0)[0]
        if len(empty_bins) > 0:
            log.warning(f"{name} has empty bins: {empty_bins}")

        # check for ill-defined stat. unc.
        nan_pos = np.where(np.isnan(self.stdev))[0]
        if len(nan_pos) > 0:
            log.warning(f"{name} has bins with ill-defined stat. unc.: {nan_pos}")

        # check whether there are any bins with ill-defined stat. uncertainty
        # but non-empty yield, those deserve a closer look
        not_empty_but_nan = [b for b in nan_pos if b not in empty_bins]
        if len(not_empty_but_nan) > 0:
            log.warning(
                f"{name} has non-empty bins with ill-defined stat. unc.: "
                f"{not_empty_but_nan}"
            )

    def normalize_to_yield(self, reference_histogram: H) -> float:
        """Normalizes a histogram to match the yield of a reference.

        Returns the normalization factor used to normalize the histogram.

        Args:
            reference_histogram (cabinetry.histo.Histogram): reference histogram to
                normalize to

        Returns:
            float: the yield ratio: un-normalized yield / normalized yield
        """
        target_integrated_yield = sum(reference_histogram.yields)
        current_integrated_yield = sum(self.yields)
        normalization_ratio = current_integrated_yield / target_integrated_yield
        # scale integrated yield to match target (also scale stdev accordingly)
        self /= normalization_ratio
        return normalization_ratio


def name(
    region: Dict[str, Any],
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    *,
    template: Optional[Literal["Up", "Down"]] = None,
) -> str:
    """Returns a unique name for each histogram.

    If the template is not None, the systematic is required to have a name (guaranteed
    as long as it follows the config schema).

    Args:
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (Optional[Literal["Up", "Down"]], optional): which template to
            consider: "Up", "Down", None for the nominal case, defaults to None

    Returns:
        str: unique name for the histogram
    """
    name = f"{region['Name']}_{sample['Name']}"
    if template is not None:
        name += f"_{systematic['Name']}_{template}"
    name = name.replace(" ", "-")
    return name



"""Collects required template histograms provided by user."""

import logging
import pathlib
from typing import Any, cast, Dict, Literal, Optional

from cabinetry import histo
from cabinetry import route
from cabinetry.templates import utils


log = logging.getLogger(__name__)


def _histo_path(
    general_path: str,
    variation_path: str,
    region: Dict[str, Any],
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    template: Optional[Literal["Up", "Down"]],
) -> str:
    """Returns the paths to a histogram for a region-sample-systematic-template.

    The path should contain a colon to separate the path to the file itself and the
    location of the histogram within the file. Due to the presence of this colon, the
    return value is a string instead of a pathlib path. A path is built starting from
    the path specified in the general options in the configuration file. This path
    contains placeholders for region-, sample-, and systematic-specific values. For
    non-nominal templates there are overrides for ``SamplePath`` and ``VariationPath``
    which can be used.

    Args:
        general_path (str): path specified in general settings, with sections that can
            be overridden by region / sample / systematic settings
        variation_path (str): default for VariationPath setting
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (Optional[Literal["Up", "Down"]]): template considered: "Up", "Down",
            or None for nominal

    Raises:
        ValueError: when ``RegionPath`` placeholder is used, but region setting is not
            specified
        ValueError: when ``SamplePath`` placeholder is used, but sample setting is not
            specified

    Returns:
        str: path to input file (and histogram within)
    """
    # obtain region and sample paths, if they are defined
    region_path = region.get("RegionPath", None)
    sample_path = sample.get("SamplePath", None)

    # check whether a systematic is being processed
    if template is not None:
        # determine whether the template has an override for SamplePath specified
        sample_override = utils._check_for_override(systematic, template, "SamplePath")
        if sample_override is not None:
            sample_path = sample_override

        # check for VariationPath override
        variation_override = utils._check_for_override(
            systematic, template, "VariationPath"
        )
        if variation_override is not None:
            # _check_for_override should return Optional[str] for VariationPath, but
            # mypy cannot know that it will not be a list, so explicitly cast to str
            variation_path = cast(str, variation_override)

    # create a new string for path handling, with placeholders replaced subsequently
    path = general_path

    # handle region-specific setting
    region_template_exists = "{RegionPath}" in path
    if region_path is not None:
        if not region_template_exists:
            log.warning(
                "region override specified, but {RegionPath} not found in default path"
            )
        path = path.replace("{RegionPath}", region_path)
    elif region_template_exists:
        raise ValueError(f"no path setting found for region {region['Name']}")

    # handle sample-specific setting
    sample_template_exists = "{SamplePath}" in path
    if sample_path is not None:
        if not sample_template_exists:
            log.warning(
                "sample override specified, but {SamplePath} not found in default path"
            )
        path = path.replace("{SamplePath}", sample_path)
    elif sample_template_exists:
        raise ValueError(f"no path setting found for sample {sample['Name']}")

    # handle variation-specific setting (variation_path is always specified via function
    # argument and possibly also via override)
    if "{VariationPath}" not in path:
        log.warning(
            "variation override specified, but {VariationPath} not found in default "
            "path"
        )
    path = path.replace("{VariationPath}", variation_path)

    # check for presence of colon to distinguish path to file and location within file
    if ":" not in path:
        log.warning(f"no colon found in path {path}, may not be able to find histogram")

    return path


def _collector(
    histogram_folder: pathlib.Path, general_path: str, variation_path: str, method: str
) -> route.ProcessorFunc:
    """Returns the histogram-collecting function to be applied to template histograms.

    Needed by ``cabinetry.route.apply_to_all_templates``. Could alternatively create a
    ``Collector`` class that contains processors (see ``builder._Builder`` for an
    example).

    Args:
        histogram_folder (pathlib.Path): folder to save the histograms to
        general_path (str): template for paths to input files for histogram building
        variation_path (str): default for VariationPath setting
        method (str): backend to use for histogram production

    Raises:
            NotImplementedError: when requesting an unknown backend

    Returns:
        route.ProcessorFunc: function to apply to a template histogram
    """

    def collect_template(
        region: Dict[str, Any],
        sample: Dict[str, Any],
        systematic: Dict[str, Any],
        template: Optional[Literal["Up", "Down"]],
    ) -> None:
        """Collects a histogram and writes it to a file.

        Args:
            region (Dict[str, Any]): containing all region information
            sample (Dict[str, Any]): containing all sample information
            systematic (Dict[str, Any]): containing all systematic information
            template (Optional[Literal["Up", "Down"]]): template considered: "Up",
                "Down", or None for nominal
        """
        histo_path = _histo_path(
            general_path, variation_path, region, sample, systematic, template
        )

        # obtain the histogram
        if method == "uproot":
            from cabinetry.contrib import histogram_reader

            histogram = histogram_reader.with_uproot(histo_path)

        else:
            raise NotImplementedError(f"unknown backend {method}")

        # store information in a Histogram instance and save it
        utils._name_and_save(
            histogram_folder,
            histo.Histogram(histogram),
            region,
            sample,
            systematic,
            template,
        )

    return collect_template



import logging
from unittest import mock

import pyhf
import pytest

from cabinetry import model_utils
from cabinetry import tabulate


@pytest.mark.parametrize(
    "test_input, expected", [(("abc", 0), "abc\nbin 1"), (("abc", 2), "abc\nbin 3")]
)
def test__header_name(test_input, expected):
    assert tabulate._header_name(*test_input) == expected
    assert tabulate._header_name("abc", 2, unique=False) == "\nbin 3"


def test__yields_per_bin(example_spec_multibin, example_spec_with_background, caplog):
    caplog.set_level(logging.DEBUG)

    # multiple channels
    model = pyhf.Workspace(example_spec_multibin).model()
    yields = [[[25.0, 5.0]], [[8.0]]]
    total_stdev = [[[5.0, 2.0], [5.0, 2.0]], [[1.0], [1.0]]]
    data = [[35, 8], [10]]
    channels = model.config.channels
    label = "pre-fit"
    caplog.clear()

    yield_table = tabulate._yields_per_bin(
        model, yields, total_stdev, data, channels, label
    )
    assert yield_table == [
        {
            "sample": "Signal",
            "region_1\nbin 1": "25.00 \u00B1 5.00",
            "region_1\nbin 2": "5.00 \u00B1 2.00",
            "region_2\nbin 1": "8.00 \u00B1 1.00",
        },
        {
            "sample": "total",
            "region_1\nbin 1": "25.00 \u00B1 5.00",
            "region_1\nbin 2": "5.00 \u00B1 2.00",
            "region_2\nbin 1": "8.00 \u00B1 1.00",
        },
        {
            "sample": "data",
            "region_1\nbin 1": "35.00",
            "region_1\nbin 2": "8.00",
            "region_2\nbin 1": "10.00",
        },
    ]
    assert "yields per bin for pre-fit model prediction:" in caplog.records[0].message
    caplog.clear()

    # multiple samples
    model = pyhf.Workspace(example_spec_with_background).model()
    yields = [[[150.0], [50.0]]]
    total_stdev = [[[6.45], [2.15], [8.60]]]
    data = [[160]]
    channels = model.config.channels

    yield_table = tabulate._yields_per_bin(
        model, yields, total_stdev, data, channels, label
    )
    assert yield_table == [
        {"sample": "Background", "Signal Region\nbin 1": "150.00 \u00B1 6.45"},
        {"sample": "Signal", "Signal Region\nbin 1": "50.00 \u00B1 2.15"},
        {"sample": "total", "Signal Region\nbin 1": "200.00 \u00B1 8.60"},
        {"sample": "data", "Signal Region\nbin 1": "160.00"},
    ]


def test__save_tables(tmp_path):
    table_dict = {
        "yields_per_bin": [
            {
                "sample": "Background",
                "Signal_region\nbin 1": "111.00 \u00B1 9.50",
                "Signal_region\nbin 2": "116.00 \u00B1 10.00",
            },
            {
                "sample": "Signal",
                "Signal_region\nbin 1": "1.00 \u00B1 0.50",
                "Signal_region\nbin 2": "5.00 \u00B1 1.00",
            },
            {
                "sample": "total",
                "Signal_region\nbin 1": "112.00 \u00B1 10.00",
                "Signal_region\nbin 2": "121.00 \u00B1 11.00",
            },
            {
                "sample": "data",
                "Signal_region\nbin 1": "112.00",
                "Signal_region\nbin 2": "123.00",
            },
        ],
        "yields_per_channel": [
            {"sample": "Background", "Signal_region": "227.00 \u00B1 16.00"},
            {"sample": "Signal", "Signal_region": "6.00 \u00B1 1.00"},
            {"sample": "total", "Signal_region": "233.00 \u00B1 17.00"},
            {"sample": "data", "Signal_region": "235.00"},
        ],
    }

    tabulate._save_tables(table_dict, tmp_path, "abc", "simple")
    fname_bin = tmp_path / "yields_per_bin_abc.txt"
    fname_channel = tmp_path / "yields_per_channel_abc.txt"
    assert fname_bin.is_file()  # bin table saved
    assert fname_channel.is_file()  # channel table saved
    assert fname_bin.read_text() == (
        "sample      Signal_region    Signal_region\n"
        "            bin 1            bin 2\n"
        "----------  ---------------  ---------------\n"
        "Background  111.00 \u00B1 9.50    116.00 \u00B1 10.00\n"
        "Signal      1.00 \u00B1 0.50      5.00 \u00B1 1.00\n"
        "total       112.00 \u00B1 10.00   121.00 \u00B1 11.00\n"
        "data        112.00           123.00\n"
    )
    assert fname_channel.read_text() == (
        "sample      Signal_region\n"
        "----------  ---------------\n"
        "Background  227.00 \u00B1 16.00\n"
        "Signal      6.00 \u00B1 1.00\n"
        "total       233.00 \u00B1 17.00\n"
        "data        235.00\n"
    )

    # suffix for latex, modification of newlines in header
    tabulate._save_tables(table_dict, tmp_path, "abc", "latex")
    fname_bin = tmp_path / "yields_per_bin_abc.tex"
    assert fname_bin.is_file()
    assert fname_bin.read_text().split("&")[1] == " Signal\\_region, bin 1   "

    # unchanged suffix for other formats
    tabulate._save_tables(table_dict, tmp_path, "abc", "html")
    fname_bin = tmp_path / "yields_per_bin_abc.html"
    assert fname_bin.is_file()


def test__yields_per_channel(
    example_spec_multibin, example_spec_with_background, caplog
):
    caplog.set_level(logging.DEBUG)

    # multiple channels
    model = pyhf.Workspace(example_spec_multibin).model()
    yields = [[30], [8.0]]
    total_stdev = [[5.39, 5.39], [1.0, 1.0]]
    data = [43, 10]
    channels = model.config.channels
    label = "pre-fit"
    caplog.clear()

    yield_table = tabulate._yields_per_channel(
        model, yields, total_stdev, data, channels, label
    )
    assert yield_table == [
        {
            "sample": "Signal",
            "region_1": "30.00 \u00B1 5.39",
            "region_2": "8.00 \u00B1 1.00",
        },
        {
            "sample": "total",
            "region_1": "30.00 \u00B1 5.39",
            "region_2": "8.00 \u00B1 1.00",
        },
        {"sample": "data", "region_1": "43.00", "region_2": "10.00"},
    ]
    assert (
        "yields per channel for pre-fit model prediction:" in caplog.records[0].message
    )
    caplog.clear()

    # multiple samples
    model = pyhf.Workspace(example_spec_with_background).model()
    yields = [[150.0, 50]]
    total_stdev = [[6.45, 2.15, 8.60]]
    data = [160]
    channels = model.config.channels

    yield_table = tabulate._yields_per_channel(
        model, yields, total_stdev, data, channels, label
    )
    assert yield_table == [
        {"sample": "Background", "Signal Region": "150.00 \u00B1 6.45"},
        {"sample": "Signal", "Signal Region": "50.00 \u00B1 2.15"},
        {"sample": "total", "Signal Region": "200.00 \u00B1 8.60"},
        {"sample": "data", "Signal Region": "160.00"},
    ]


@mock.patch("cabinetry.tabulate._save_tables")
@mock.patch(
    "cabinetry.tabulate._yields_per_channel",
    return_value=[{"yields_per_channel": None}],
)
@mock.patch(
    "cabinetry.tabulate._yields_per_bin", return_value=[{"yields_per_bin": None}]
)
@mock.patch("cabinetry.model_utils._filter_channels", side_effect=[["SR"], [], ["SR"]])
@mock.patch("cabinetry.model_utils._data_per_channel", return_value=[[12.0]])
def test_yields(
    mock_data, mock_filter, mock_bin, mock_channel, mock_save, example_spec, caplog
):
    # the return values of the functions producing the actual tables are hardcoded above
    # to be able to check that they are correctly propagated to the output
    caplog.set_level(logging.DEBUG)
    model = pyhf.Workspace(example_spec).model()
    model_pred = model_utils.ModelPrediction(
        model, [[[10.0]]], [[[0.3], [0.3]]], [[0.3, 0.3]], "pred"
    )
    data = [12.0, 1.0]  # with auxdata to strip via mock

    table_dict = tabulate.yields(model_pred, data)
    assert mock_data.call_args_list == [((model, data), {})]
    assert mock_filter.call_args_list == [((model, None), {})]
    assert mock_bin.call_args_list == [
        ((model, [[[10.0]]], [[[0.3], [0.3]]], [[12.0]], ["SR"], "pred"), {})
    ]
    assert mock_channel.call_count == 0
    assert mock_save.call_count == 1
    assert table_dict == {"yields_per_bin": [{"yields_per_bin": None}]}

    # no table to produce (does not call _filter_channels)
    table_dict = tabulate.yields(model_pred, data, per_bin=False, per_channel=False)
    assert "requested neither yields per bin nor per channel, no table produced" in [
        rec.message for rec in caplog.records
    ]
    assert table_dict == {}
    caplog.clear()

    # no channels to include
    table_dict = tabulate.yields(model_pred, data, channels="abc")
    assert mock_filter.call_args == ((model, "abc"), {})
    assert mock_bin.call_count == 1  # one call from before, no new call
    assert mock_channel.call_count == 0
    assert table_dict == {}

    # yields per channel, not per bin, do not save tables
    table_dict = tabulate.yields(
        model_pred, data, per_bin=False, per_channel=True, save_tables=False
    )
    assert mock_bin.call_count == 1  # one call from before
    assert mock_channel.call_args_list == [
        ((model, [[10.0]], [[0.3, 0.3]], [12.0], ["SR"], "pred"), {})
    ]
    assert table_dict == {"yields_per_channel": [{"yields_per_channel": None}]}
    assert mock_save.call_count == 1  # one call from before, no new call



"""Implements the command line interface."""

import io
import json
import logging
import pathlib
from typing import Any, List, Optional, Tuple

import click
import yaml

from cabinetry import __version__
from cabinetry import configuration as cabinetry_configuration
from cabinetry import fit as cabinetry_fit
from cabinetry import model_utils as cabinetry_model_utils
from cabinetry import tabulate as cabinetry_tabulate
from cabinetry import templates as cabinetry_templates
from cabinetry import visualize as cabinetry_visualize
from cabinetry import workspace as cabinetry_workspace


class OrderedGroup(click.Group):
    """A group that shows commands in the order they were added."""

    def list_commands(self, _: Any) -> List[str]:
        """Returns a list of commands."""
        return list(self.commands.keys())


def _set_logging() -> None:
    """Sets log levels and format for CLI."""
    logging.basicConfig(
        level=logging.INFO, format="%(levelname)s - %(name)s - %(message)s"
    )
    logging.getLogger("pyhf").setLevel(logging.WARNING)


@click.version_option(version=__version__)
@click.group(cls=OrderedGroup)
def cabinetry() -> None:
    """Entrypoint to the cabinetry CLI."""


@click.command()
@click.argument("config", type=click.File("r"))
@click.option(
    "--method",
    default="uproot",
    help="backend for histogram production (default: uproot)",
)
def templates(config: io.TextIOWrapper, method: str) -> None:
    """Produces template histograms.

    CONFIG: path to cabinetry configuration file
    """
    _set_logging()
    cabinetry_config = yaml.safe_load(config)
    cabinetry_configuration.validate(cabinetry_config)
    cabinetry_templates.build(cabinetry_config, method=method)


@click.command()
@click.argument("config", type=click.File("r"))
def postprocess(config: io.TextIOWrapper) -> None:
    """Post-processes template histograms.

    CONFIG: path to cabinetry configuration file
    """
    _set_logging()
    cabinetry_config = yaml.safe_load(config)
    cabinetry_configuration.validate(cabinetry_config)
    cabinetry_templates.postprocess(cabinetry_config)


@click.command()
@click.argument("config", type=click.File("r"))
@click.argument("ws_spec", type=click.File("w"))
def workspace(config: io.TextIOWrapper, ws_spec: click.utils.LazyFile) -> None:
    """Produces a ``pyhf`` workspace.

    CONFIG: path to cabinetry configuration file

    WS_SPEC: where to save the workspace containing the fit model
    """
    _set_logging()
    cabinetry_config = yaml.safe_load(config)
    cabinetry_configuration.validate(cabinetry_config)
    ws = cabinetry_workspace.build(cabinetry_config)
    # create folder containing workspace if needed
    pathlib.Path(ws_spec.name).parent.mkdir(parents=True, exist_ok=True)
    ws_spec.write(json.dumps(ws, sort_keys=True, indent=4))


@click.command()
@click.argument("ws_spec", type=click.File("r"))
@click.option("--asimov", is_flag=True, help="fit Asimov dataset (default: False)")
@click.option(
    "--minos",
    type=str,
    multiple=True,
    help="run MINOS for a parameter (default: disabled)",
)
@click.option(
    "--goodness_of_fit", is_flag=True, help="calculate goodness-of-fit (default: False)"
)
@click.option("--pulls", is_flag=True, help="produce pull plot (default: False)")
@click.option(
    "--corrmat", is_flag=True, help="produce correlation matrix (default: False)"
)
@click.option(
    "--figfolder",
    default="figures",
    help='folder to save figures to (default: "figures")',
)
def fit(
    ws_spec: io.TextIOWrapper,
    asimov: bool,
    minos: Tuple[str, ...],
    goodness_of_fit: bool,
    pulls: bool,
    corrmat: bool,
    figfolder: str,
) -> None:
    """Fits a workspace and optionally visualizes the results.

    WS_SPEC: path to workspace used in fit
    """
    _set_logging()
    # convert minos argument to None if no parameter is specified, otherwise to a list
    if len(minos) == 0:
        minos_converted = None
    else:
        minos_converted = list(minos)
    ws = json.load(ws_spec)
    model, data = cabinetry_model_utils.model_and_data(ws, asimov=asimov)
    fit_results = cabinetry_fit.fit(
        model, data, minos=minos_converted, goodness_of_fit=goodness_of_fit
    )
    if pulls:
        cabinetry_visualize.pulls(fit_results, figure_folder=figfolder)
    if corrmat:
        cabinetry_visualize.correlation_matrix(fit_results, figure_folder=figfolder)
    pass  # fixes coverage, see https://github.com/nedbat/coveragepy/issues/198


@click.command()
@click.argument("ws_spec", type=click.File("r"))
@click.option("--asimov", is_flag=True, help="fit Asimov dataset (default: False)")
@click.option(
    "--max_pars", default=10, help="maximum amount of parameters in plot (default: 10)"
)
@click.option(
    "--figfolder",
    default="figures",
    help='folder to save figures to (default: "figures")',
)
def ranking(
    ws_spec: io.TextIOWrapper, asimov: bool, max_pars: int, figfolder: str
) -> None:
    """Ranks nuisance parameters and visualizes the result.

    WS_SPEC: path to workspace used in fit
    """
    _set_logging()
    ws = json.load(ws_spec)
    model, data = cabinetry_model_utils.model_and_data(ws, asimov=asimov)
    fit_results = cabinetry_fit.fit(model, data)
    ranking_results = cabinetry_fit.ranking(model, data, fit_results=fit_results)
    cabinetry_visualize.ranking(
        ranking_results, figure_folder=figfolder, max_pars=max_pars
    )


@click.command()
@click.argument("ws_spec", type=click.File("r"))
@click.argument("par_name", type=str)
@click.option(
    "--lower_bound",
    default=None,
    type=float,
    help="lower parameter bound in scan (default: auto)",
)
@click.option(
    "--upper_bound",
    default=None,
    type=float,
    help="upper parameter bound in scan (default: auto)",
)
@click.option("--n_steps", default=11, help="number of steps in scan (default: 11)")
@click.option("--asimov", is_flag=True, help="fit Asimov dataset (default: False)")
@click.option(
    "--figfolder",
    default="figures",
    help='folder to save figures to (default: "figures")',
)
def scan(
    ws_spec: io.TextIOWrapper,
    par_name: str,
    lower_bound: Optional[float],
    upper_bound: Optional[float],
    n_steps: int,
    asimov: bool,
    figfolder: str,
) -> None:
    """Performs and visualizes a likelihood scan over a parameter.

    Parameter bounds are determined automatically, unless both the ``lower_bound`` and
    ``upper_bound`` parameters are provided.

    WS_SPEC: path to workspace used in fit

    PAR_NAME: name of parameter to scan over
    """
    _set_logging()
    par_range: Optional[Tuple[float, float]]
    if (lower_bound is not None) and (upper_bound is not None):
        # both bounds specified
        par_range = (lower_bound, upper_bound)
    elif (lower_bound is not None) or (upper_bound is not None):
        # mixed case not supported
        raise ValueError(
            "Need to either specify both lower_bound and upper_bound, or neither."
        )
    else:
        # no bounds specified
        par_range = None

    ws = json.load(ws_spec)
    model, data = cabinetry_model_utils.model_and_data(ws, asimov=asimov)
    scan_results = cabinetry_fit.scan(
        model, data, par_name, par_range=par_range, n_steps=n_steps
    )
    cabinetry_visualize.scan(scan_results, figure_folder=figfolder)


@click.command()
@click.argument("ws_spec", type=click.File("r"))
@click.option("--asimov", is_flag=True, help="fit Asimov dataset (default: False)")
@click.option(
    "--tolerance",
    default=0.01,
    help="tolerance for convergence to CLs=1-confidence_level (default: 0.01)",
)
@click.option(
    "--confidence_level",
    "--cl",
    default=0.95,
    type=click.FloatRange(0.0, 1.0, min_open=True, max_open=True),
    help="confidence level for parameter limits (default: 0.95)",
)
@click.option(
    "--figfolder",
    default="figures",
    help='folder to save figures to (default: "figures")',
)
def limit(
    ws_spec: io.TextIOWrapper,
    asimov: bool,
    tolerance: float,
    confidence_level: float,
    figfolder: str,
) -> None:
    """Calculates upper limits and visualizes CLs distribution.

    WS_SPEC: path to workspace used in fit
    """
    _set_logging()
    ws = json.load(ws_spec)
    model, data = cabinetry_model_utils.model_and_data(ws, asimov=asimov)
    limit_results = cabinetry_fit.limit(
        model, data, tolerance=tolerance, confidence_level=confidence_level
    )
    cabinetry_visualize.limit(limit_results, figure_folder=figfolder)


@click.command()
@click.argument("ws_spec", type=click.File("r"))
@click.option("--asimov", is_flag=True, help="fit Asimov dataset (default: False)")
def significance(ws_spec: io.TextIOWrapper, asimov: bool) -> None:
    """Calculates observed and expected discovery significance.

    WS_SPEC: path to workspace used in fit
    """
    _set_logging()
    ws = json.load(ws_spec)
    model, data = cabinetry_model_utils.model_and_data(ws, asimov=asimov)
    _ = cabinetry_fit.significance(model, data)


@click.command()
@click.argument("ws_spec", type=click.File("r"))
@click.option("--config", type=click.File("r"), help="cabinetry configuration file")
@click.option(
    "--postfit", is_flag=True, help="visualize post-fit model (default: pre-fit model)"
)
@click.option(
    "--figfolder",
    default="figures",
    help='folder to save figures to (default: "figures")',
)
def data_mc(
    ws_spec: io.TextIOWrapper,
    config: Optional[io.TextIOWrapper],
    postfit: bool,
    figfolder: str,
) -> None:
    """Visualizes distributions of fit model and observed data.

    WS_SPEC: path to workspace
    """
    _set_logging()
    ws = json.load(ws_spec)
    model, data = cabinetry_model_utils.model_and_data(ws)

    if config is not None:
        cabinetry_config = yaml.safe_load(config)
        cabinetry_configuration.validate(cabinetry_config)
    else:
        cabinetry_config = None

    # optionally perform maximum likelihood fit to obtain post-fit model
    fit_results = cabinetry_fit.fit(model, data) if postfit else None

    model_prediction = cabinetry_model_utils.prediction(model, fit_results=fit_results)
    cabinetry_visualize.data_mc(
        model_prediction,
        data,
        config=cabinetry_config,
        figure_folder=figfolder,
        close_figure=True,
        save_figure=True,
    )


@click.command()
@click.argument("ws_spec", type=click.File("r"))
@click.option(
    "--postfit", is_flag=True, help="show post-fit model (default: pre-fit model)"
)
@click.option(
    "--tablefolder",
    default="tables",
    help='folder to save tables to (default: "tables")',
)
@click.option(
    "--tablefmt",
    default="simple",
    help='format in which to save the table (default: "simple")',
)
def yields(
    ws_spec: io.TextIOWrapper,
    postfit: bool,
    tablefolder: str,
    tablefmt: str,
) -> None:
    """Creates yield tables of fit model and observed data.

    WS_SPEC: path to workspace
    """
    _set_logging()
    ws = json.load(ws_spec)
    model, data = cabinetry_model_utils.model_and_data(ws)

    # optionally perform maximum likelihood fit to obtain post-fit model
    fit_results = cabinetry_fit.fit(model, data) if postfit else None

    model_prediction = cabinetry_model_utils.prediction(model, fit_results=fit_results)
    cabinetry_tabulate.yields(
        model_prediction,
        data,
        table_folder=tablefolder,
        table_format=tablefmt,
    )


@click.command()
@click.argument("ws_spec", type=click.File("r"))
@click.option(
    "--split_by_sample",
    is_flag=True,
    help="split grids by sample (default: split by channel)",
)
@click.option(
    "--figfolder",
    default="figures",
    help='folder to save figures to (default: "figures")',
)
def modifier_grid(
    ws_spec: io.TextIOWrapper, split_by_sample: bool, figfolder: str
) -> None:
    """Visualizes modifier structure of a model.

    WS_SPEC: path to workspace
    """
    _set_logging()
    ws = json.load(ws_spec)
    model, _ = cabinetry_model_utils.model_and_data(ws)
    cabinetry_visualize.modifier_grid(
        model,
        figure_folder=figfolder,
        split_by_sample=split_by_sample,
        close_figure=True,
        save_figure=True,
    )


cabinetry.add_command(templates)
cabinetry.add_command(postprocess)
cabinetry.add_command(workspace)
cabinetry.add_command(fit)
cabinetry.add_command(ranking)
cabinetry.add_command(scan)
cabinetry.add_command(limit)
cabinetry.add_command(significance)
cabinetry.add_command(data_mc)
cabinetry.add_command(yields)
cabinetry.add_command(modifier_grid)



import logging
import pathlib
from unittest import mock

import numpy as np
import pytest

from cabinetry import histo
from cabinetry.templates import postprocessor


@pytest.mark.parametrize(
    "test_histo, fixed_stdev",
    [
        (
            histo.Histogram.from_arrays([1, 2, 3], [1, 2], [float("nan"), 0.2]),
            [0.0, 0.2],
        ),
        (histo.Histogram.from_arrays([1, 2, 3], [1, 2], [0.1, 0.2]), [0.1, 0.2]),
    ],
)
def test__fix_stat_unc(test_histo, fixed_stdev):
    name = "test_histo"
    postprocessor._fix_stat_unc(test_histo, name)
    assert np.allclose(test_histo.stdev, fixed_stdev)


@mock.patch("cabinetry.smooth.smooth_353qh_twice", return_value=np.asarray([1, 1.3]))
def test__apply_353qh_twice(mock_smooth):
    var = histo.Histogram.from_arrays([1, 2, 3], [1, 1.5], [0.1, 0.1])
    nom = histo.Histogram.from_arrays([1, 2, 3], [1, 1.2], [0.1, 0.1])

    postprocessor._apply_353qh_twice(var, nom, "abc")

    assert np.allclose(nom.yields, [1, 1.2])  # nominal unchanged
    # result of smoothing is [1, 1.3], multiplied by [1, 1.2] -> [1, 1.56]
    # [1, 1.56] scaled to integral 2.5 results in [0.9765625, 1.5234375]
    assert np.allclose(var.yields, [0.9765625, 1.5234375])


def test__smoothing_algorithm():
    reg = {"Name": "reg"}
    sam = {"Name": "sam"}
    sys = {"Name": "sys"}

    assert postprocessor._smoothing_algorithm(reg, sam, sys) is None

    sys = {"Name": "sys", "Smoothing": {"Algorithm": "abc"}}
    assert postprocessor._smoothing_algorithm(reg, sam, sys) == "abc"

    # region is missing in setting
    sys = {"Name": "sys", "Smoothing": {"Algorithm": "abc", "Regions": "r"}}
    assert postprocessor._smoothing_algorithm(reg, sam, sys) is None

    # region is included in setting
    sys = {"Name": "sys", "Smoothing": {"Algorithm": "abc", "Regions": ["r", "reg"]}}
    assert postprocessor._smoothing_algorithm(reg, sam, sys) == "abc"

    # sample is missing in setting
    sys = {"Name": "sys", "Smoothing": {"Algorithm": "abc", "Samples": "s"}}
    assert postprocessor._smoothing_algorithm(reg, sam, sys) is None

    # sample is included in setting
    sys = {"Name": "sys", "Smoothing": {"Algorithm": "abc", "Samples": ["s", "sam"]}}
    assert postprocessor._smoothing_algorithm(reg, sam, sys) == "abc"


@mock.patch("cabinetry.templates.postprocessor._apply_353qh_twice")
@mock.patch("cabinetry.templates.postprocessor._fix_stat_unc")
def test_apply_postprocessing(mock_stat, mock_smooth, caplog):
    caplog.set_level(logging.DEBUG)
    histogram = histo.Histogram.from_arrays([1, 2, 3], [1, 1], [float("nan"), 0.2])
    nom_hist = histo.Histogram.from_arrays([1, 2, 3], [2, 2], [0.1, 0.2])
    name = "test_histo"
    modified_histogram = postprocessor.apply_postprocessing(
        histogram, name, smoothing_algorithm="353QH, twice", nominal_histogram=nom_hist
    )

    # call to stat. unc. fix
    assert mock_stat.call_count == 1
    assert np.allclose(mock_stat.call_args[0][0].yields, histogram.yields)
    assert np.allclose(mock_stat.call_args[0][0].stdev, histogram.stdev, equal_nan=True)
    assert mock_stat.call_args[0][1] == name
    assert mock_stat.call_args[1] == {}

    # call to smoothing
    assert mock_smooth.call_count == 1
    assert np.allclose(mock_smooth.call_args[0][0].yields, histogram.yields)
    assert np.allclose(
        mock_smooth.call_args[0][0].stdev, histogram.stdev, equal_nan=True
    )
    assert np.allclose(mock_smooth.call_args[0][1].yields, nom_hist.yields)
    assert np.allclose(mock_smooth.call_args[0][1].stdev, nom_hist.stdev)
    assert mock_smooth.call_args[0][2] == name
    assert mock_smooth.call_args[1] == {}

    # the original histogram should be unchanged
    assert np.allclose(histogram.yields, [1, 1])
    assert np.allclose(histogram.stdev, [float("nan"), 0.2], equal_nan=True)
    # original and modified histogram should match (modifications were mocked out)
    assert np.allclose(histogram.yields, modified_histogram.yields)
    assert np.allclose(histogram.stdev, modified_histogram.stdev, equal_nan=True)

    # unknown smoothing algorithm
    _ = postprocessor.apply_postprocessing(
        histogram, name, smoothing_algorithm="abc", nominal_histogram=nom_hist
    )
    assert mock_stat.call_count == 2
    assert mock_smooth.call_count == 1
    assert "unknown smoothing algorithm abc" in [rec.message for rec in caplog.records]
    caplog.clear()

    # no smoothing
    _ = postprocessor.apply_postprocessing(
        histogram, name, smoothing_algorithm=None, nominal_histogram=None
    )
    assert mock_stat.call_count == 3
    assert mock_smooth.call_count == 1

    # known smoothing, but no nominal histogram
    with pytest.raises(
        ValueError, match="cannot apply smoothing, nominal histogram missing"
    ):
        _ = postprocessor.apply_postprocessing(
            histogram, name, smoothing_algorithm="353QH, twice", nominal_histogram=None
        )


@mock.patch("cabinetry.histo.name", return_value="histo_name")
def test__postprocessor(mock_name):
    processor = postprocessor._postprocessor(pathlib.Path("path"))

    region = {"Name": "region"}
    sample = {"Name": "sample"}
    systematic = {"Name": "systematic"}
    template = "Up"

    mock_original_histogram = mock.MagicMock()
    mock_new_histogram = mock.MagicMock()
    with mock.patch(
        "cabinetry.histo.Histogram.from_config", return_value=mock_original_histogram
    ) as mock_from_config:
        with mock.patch(
            "cabinetry.templates.postprocessor.apply_postprocessing",
            return_value=mock_new_histogram,
        ) as mock_postprocessing:
            # execute the provided function
            with mock.patch(
                "cabinetry.templates.postprocessor._smoothing_algorithm",
                return_value=None,
            ) as mock_smooth:
                processor(region, sample, systematic, template)
                assert mock_smooth.call_args_list == [
                    ((region, sample, systematic), {})
                ]

            # check that the relevant functions were called
            assert mock_from_config.call_args_list == [
                (
                    (pathlib.Path("path"), region, sample, systematic),
                    {"template": template, "modified": False},
                )
            ]  # histogram was created

            assert mock_postprocessing.call_args_list == [
                (
                    (mock_original_histogram, "histo_name"),
                    {"smoothing_algorithm": None, "nominal_histogram": None},
                )
            ]  # postprocessing was executed

            assert mock_new_histogram.save.call_args_list == [
                ((pathlib.Path("path/histo_name_modified"),), {})
            ]  # new histogram was saved (meaning right name was obtained)

            # postprocessor with smoothing
            systematic = {
                "Name": "systematic",
                "Smoothing": {"Algorithm": "353QH, twice"},
            }
            with mock.patch(
                "cabinetry.templates.postprocessor._smoothing_algorithm",
                return_value="353QH, twice",
            ) as mock_smooth:
                processor(region, sample, systematic, template)
                assert mock_smooth.call_args_list == [
                    ((region, sample, systematic), {})
                ]

            # nominal histogram was read
            assert mock_from_config.call_args[0][3] == {}

            # postprocessing called with smoothing algorithm and nominal histogram
            assert mock_postprocessing.call_args == (
                (mock_original_histogram, "histo_name"),
                {
                    "smoothing_algorithm": "353QH, twice",
                    "nominal_histogram": mock_original_histogram,
                },
            )



import pathlib
from unittest import mock

from cabinetry.templates import utils


def test__check_for_override():
    # override exists for template
    assert (
        utils._check_for_override({"Up": {"setting": "val"}}, "Up", "setting") == "val"
    )

    # no override for template
    assert utils._check_for_override({}, "Up", "setting") is None

    # no option requested
    assert utils._check_for_override({"Up": {"setting": "val"}}, "Up", "") is None

    # override is a list
    assert utils._check_for_override(
        {"Up": {"setting": ["val", "val2"]}}, "Up", "setting"
    ) == ["val", "val2"]


@mock.patch("cabinetry.histo.name", return_value="name")
def test__name_and_save(mock_name):
    histogram_folder = pathlib.Path("path")
    histogram = mock.MagicMock()
    region = {"Name": "test_region"}
    sample = {"Name": "sample"}
    systematic = {}

    utils._name_and_save(histogram_folder, histogram, region, sample, systematic, "Up")

    # check that the naming function was called, the histogram was validated and saved
    assert mock_name.call_args_list == [
        ((region, sample, systematic), {"template": "Up"})
    ]
    assert histogram.validate.call_args_list == [mock.call("name")]
    assert histogram.save.call_args_list == [mock.call(pathlib.Path("path/name"))]



import json
import pathlib
from unittest import mock

from click.testing import CliRunner
import numpy as np
import pytest
import yaml

from cabinetry import __version__
from cabinetry import cli
from cabinetry import fit


class CLIHelpers:
    @staticmethod
    def write_config(path, config):
        with open(path, "w") as f:
            yaml.dump(config, f)


@pytest.fixture
def cli_helpers():
    return CLIHelpers


def test_cabinetry():
    runner = CliRunner()
    result = runner.invoke(cli.cabinetry, ["--help"])
    assert result.exit_code == 0
    assert "Entrypoint to the cabinetry CLI." in result.output

    result = runner.invoke(cli.cabinetry, ["--version"])
    assert result.exit_code == 0
    assert f"cabinetry, version {__version__}" in result.output


# using autospec to catch changes in public API
@mock.patch("cabinetry.templates.build", autospec=True)
@mock.patch("cabinetry.configuration.validate", autospec=True)
def test_templates(mock_validate, mock_create_histograms, cli_helpers, tmp_path):
    config = {"General": {"Measurement": "test_config"}}

    config_path = str(tmp_path / "config.yml")
    cli_helpers.write_config(config_path, config)

    runner = CliRunner()

    # default method
    result = runner.invoke(cli.templates, [config_path])
    assert result.exit_code == 0
    assert mock_validate.call_args_list == [((config,), {})]
    assert mock_create_histograms.call_args_list == [((config,), {"method": "uproot"})]

    # different method
    result = runner.invoke(cli.templates, ["--method", "unknown", config_path])
    assert result.exit_code == 0
    assert mock_create_histograms.call_args == ((config,), {"method": "unknown"})


@mock.patch("cabinetry.templates.postprocess", autospec=True)
@mock.patch("cabinetry.configuration.validate", autospec=True)
def test_postprocess(mock_validate, mock_postprocess, cli_helpers, tmp_path):
    config = {"General": {"Measurement": "test_config"}}

    config_path = str(tmp_path / "config.yml")
    cli_helpers.write_config(config_path, config)

    runner = CliRunner()

    result = runner.invoke(cli.postprocess, [config_path])
    assert result.exit_code == 0
    assert mock_validate.call_args_list == [((config,), {})]
    assert mock_postprocess.call_args_list == [((config,), {})]


@mock.patch(
    "cabinetry.workspace.build", return_value={"workspace": "mock"}, autospec=True
)
@mock.patch("cabinetry.configuration.validate", autospec=True)
def test_workspace(mock_validate, mock_build, cli_helpers, tmp_path):
    config = {"General": {"Measurement": "test_config"}}

    config_path = str(tmp_path / "config.yml")
    cli_helpers.write_config(config_path, config)

    workspace_path = str(tmp_path / "workspace.json")

    runner = CliRunner()

    result = runner.invoke(cli.workspace, [config_path, workspace_path])
    assert result.exit_code == 0
    assert mock_validate.call_args_list == [((config,), {})]
    assert mock_build.call_args_list == [((config,), {})]
    assert json.loads(pathlib.Path(workspace_path).read_text()) == {"workspace": "mock"}


@mock.patch("cabinetry.visualize.correlation_matrix", autospec=True)
@mock.patch("cabinetry.visualize.pulls", autospec=True)
@mock.patch(
    "cabinetry.fit.fit",
    return_value=fit.FitResults(
        np.asarray([1.0]), np.asarray([0.1]), ["label"], np.asarray([[1.0]]), 1.0
    ),
    autospec=True,
)
@mock.patch(
    "cabinetry.model_utils.model_and_data",
    return_value=("model", "data"),
    autospec=True,
)
def test_fit(mock_utils, mock_fit, mock_pulls, mock_corrmat, tmp_path):
    workspace = {"workspace": "mock"}
    bestfit = np.asarray([1.0])
    uncertainty = np.asarray([0.1])
    labels = ["label"]
    corr_mat = np.asarray([[1.0]])
    fit_results = fit.FitResults(bestfit, uncertainty, labels, corr_mat, 1.0)

    workspace_path = str(tmp_path / "workspace.json")

    # need to save workspace to file since click looks for it
    with open(workspace_path, "w") as f:
        f.write('{"workspace": "mock"}')

    runner = CliRunner()

    # default
    result = runner.invoke(cli.fit, [workspace_path])
    assert result.exit_code == 0
    assert mock_utils.call_args_list == [((workspace,), {"asimov": False})]
    assert mock_fit.call_args_list == [
        (("model", "data"), {"minos": None, "goodness_of_fit": False})
    ]

    # Asimov
    result = runner.invoke(cli.fit, ["--asimov", workspace_path])
    assert result.exit_code == 0
    assert mock_utils.call_args == ((workspace,), {"asimov": True})
    assert mock_fit.call_args == (
        ("model", "data"),
        {"minos": None, "goodness_of_fit": False},
    )

    # MINOS for one parameter
    result = runner.invoke(cli.fit, ["--minos", "par", workspace_path])
    assert result.exit_code == 0
    assert mock_utils.call_args == ((workspace,), {"asimov": False})
    assert mock_fit.call_args == (
        ("model", "data"),
        {"minos": ["par"], "goodness_of_fit": False},
    )

    # MINOS for multiple parameters
    result = runner.invoke(
        cli.fit, ["--minos", "par_a", "--minos", "par_b", workspace_path]
    )
    assert result.exit_code == 0
    assert mock_utils.call_args == ((workspace,), {"asimov": False})
    assert mock_fit.call_args == (
        ("model", "data"),
        {"minos": ["par_a", "par_b"], "goodness_of_fit": False},
    )

    # goodness-of-fit
    result = runner.invoke(cli.fit, ["--goodness_of_fit", workspace_path])
    assert result.exit_code == 0
    assert mock_utils.call_args == ((workspace,), {"asimov": False})
    assert mock_fit.call_args == (
        ("model", "data"),
        {"minos": None, "goodness_of_fit": True},
    )

    # pull plot
    result = runner.invoke(cli.fit, ["--pulls", workspace_path])
    assert result.exit_code == 0
    assert mock_pulls.call_args_list == [((fit_results,), {"figure_folder": "figures"})]

    # correlation matrix plot
    result = runner.invoke(cli.fit, ["--corrmat", workspace_path])
    assert result.exit_code == 0
    assert mock_corrmat.call_args_list == [
        ((fit_results,), {"figure_folder": "figures"})
    ]

    # both plots, different folder
    result = runner.invoke(
        cli.fit, ["--figfolder", "folder", "--pulls", "--corrmat", workspace_path]
    )
    assert result.exit_code == 0
    assert mock_corrmat.call_args == ((fit_results,), {"figure_folder": "folder"})
    assert mock_pulls.call_args == ((fit_results,), {"figure_folder": "folder"})


@mock.patch("cabinetry.visualize.ranking", autospec=True)
@mock.patch(
    "cabinetry.fit.ranking",
    return_value=fit.RankingResults(
        np.asarray([1.0]),
        np.asarray([0.1]),
        ["label"],
        np.asarray([[1.2]]),
        np.asarray([[0.8]]),
        np.asarray([[1.1]]),
        np.asarray([[0.9]]),
    ),
    autospec=True,
)
@mock.patch(
    "cabinetry.fit.fit",
    return_value=fit.FitResults(
        np.asarray([1.0]), np.asarray([0.1]), ["label"], np.asarray([[1.0]]), 1.0
    ),
    autospec=True,
)
@mock.patch(
    "cabinetry.model_utils.model_and_data",
    return_value=("model", "data"),
    autospec=True,
)
def test_ranking(mock_utils, mock_fit, mock_rank, mock_vis, tmp_path):
    workspace = {"workspace": "mock"}
    bestfit = np.asarray([1.0])
    uncertainty = np.asarray([0.1])
    labels = ["label"]
    corr_mat = np.asarray([[1.0]])
    fit_results = fit.FitResults(bestfit, uncertainty, labels, corr_mat, 1.0)

    workspace_path = str(tmp_path / "workspace.json")

    # need to save workspace to file since click looks for it
    with open(workspace_path, "w") as f:
        f.write('{"workspace": "mock"}')

    runner = CliRunner()

    # default
    result = runner.invoke(cli.ranking, [workspace_path])
    assert result.exit_code == 0
    assert mock_utils.call_args_list == [((workspace,), {"asimov": False})]
    assert mock_fit.call_args_list == [(("model", "data"), {})]
    assert mock_rank.call_args_list == [
        (("model", "data"), {"fit_results": fit_results})
    ]
    assert mock_vis.call_count == 1
    assert np.allclose(mock_vis.call_args[0][0].prefit_up, [[1.2]])
    assert np.allclose(mock_vis.call_args[0][0].prefit_down, [[0.8]])
    assert np.allclose(mock_vis.call_args[0][0].postfit_up, [[1.1]])
    assert np.allclose(mock_vis.call_args[0][0].postfit_down, [[0.9]])
    assert mock_vis.call_args[1] == {"figure_folder": "figures", "max_pars": 10}

    # Asimov, maximum amount of parameters, custom folder
    result = runner.invoke(
        cli.ranking,
        ["--asimov", "--max_pars", 3, "--figfolder", "folder", workspace_path],
    )
    assert result.exit_code == 0
    assert mock_utils.call_args == ((workspace,), {"asimov": True})
    assert mock_fit.call_args == (("model", "data"), {})
    assert mock_rank.call_args == (("model", "data"), {"fit_results": fit_results})
    assert mock_vis.call_args[1] == {"figure_folder": "folder", "max_pars": 3}


@mock.patch("cabinetry.visualize.scan", autospec=True)
@mock.patch(
    "cabinetry.fit.scan",
    return_value=fit.ScanResults("par", 1.0, 0.1, np.asarray([1.5]), np.asarray([3.5])),
    autospec=True,
)
@mock.patch(
    "cabinetry.model_utils.model_and_data",
    return_value=("model", "data"),
    autospec=True,
)
def test_scan(mock_utils, mock_scan, mock_vis, tmp_path):
    workspace = {"workspace": "mock"}
    workspace_path = str(tmp_path / "workspace.json")

    # need to save workspace to file since click looks for it
    with open(workspace_path, "w") as f:
        f.write('{"workspace": "mock"}')

    par_name = "par"
    scan_results = fit.ScanResults(
        par_name, 1.0, 0.1, np.asarray([1.5]), np.asarray([3.5])
    )

    runner = CliRunner()

    # default
    result = runner.invoke(cli.scan, [workspace_path, par_name])
    assert result.exit_code == 0
    assert mock_utils.call_args_list == [((workspace,), {"asimov": False})]
    assert mock_scan.call_args_list == [
        (("model", "data", par_name), {"par_range": None, "n_steps": 11})
    ]
    assert mock_vis.call_count == 1
    assert mock_vis.call_args[0][0].name == scan_results.name
    assert mock_vis.call_args[0][0].bestfit == scan_results.bestfit
    assert mock_vis.call_args[0][0].uncertainty == scan_results.uncertainty
    assert np.allclose(
        mock_vis.call_args[0][0].parameter_values, scan_results.parameter_values
    )
    assert np.allclose(mock_vis.call_args[0][0].delta_nlls, scan_results.delta_nlls)
    assert mock_vis.call_args[1] == {"figure_folder": "figures"}

    # only one bound
    with pytest.raises(
        ValueError,
        match="Need to either specify both lower_bound and upper_bound, or neither.",
    ):
        runner.invoke(
            cli.scan,
            ["--lower_bound", 1.0, workspace_path, par_name],
            catch_exceptions=False,
        )
    with pytest.raises(
        ValueError,
        match="Need to either specify both lower_bound and upper_bound, or neither.",
    ):
        runner.invoke(
            cli.scan,
            ["--upper_bound", 1.0, workspace_path, par_name],
            catch_exceptions=False,
        )

    # custom bounds, number of steps and Asimov
    result = runner.invoke(
        cli.scan,
        [
            "--lower_bound",
            0.0,
            "--upper_bound",
            2.0,
            "--n_steps",
            21,
            "--asimov",
            "--figfolder",
            "folder",
            workspace_path,
            par_name,
        ],
    )
    assert result.exit_code == 0
    assert mock_utils.call_args == ((workspace,), {"asimov": True})
    assert mock_scan.call_args == (
        ("model", "data", par_name),
        {"par_range": (0.0, 2.0), "n_steps": 21},
    )
    assert mock_vis.call_args[1] == {"figure_folder": "folder"}


@mock.patch("cabinetry.visualize.limit", autospec=True)
@mock.patch(
    "cabinetry.fit.limit",
    return_value=fit.LimitResults(
        3.0,
        np.asarray([1.0, 2.0, 3.0, 4.0, 5.0]),
        np.asarray([0.05]),
        np.asarray([0.01, 0.02, 0.05, 0.07, 0.10]),
        np.asarray([3.0]),
        0.90,
    ),
    autospec=True,
)
@mock.patch(
    "cabinetry.model_utils.model_and_data",
    return_value=("model", "data"),
    autospec=True,
)
def test_limit(mock_utils, mock_limit, mock_vis, tmp_path):
    workspace = {"workspace": "mock"}
    workspace_path = str(tmp_path / "workspace.json")

    # need to save workspace to file since click looks for it
    with open(workspace_path, "w") as f:
        f.write('{"workspace": "mock"}')

    limit_results = fit.LimitResults(
        3.0,
        np.asarray([1.0, 2.0, 3.0, 4.0, 5.0]),
        np.asarray([0.05]),
        np.asarray([0.01, 0.02, 0.05, 0.07, 0.10]),
        np.asarray([3.0]),
        0.90,
    )

    runner = CliRunner()

    # default
    result = runner.invoke(cli.limit, [workspace_path])
    assert result.exit_code == 0
    assert mock_utils.call_args_list == [((workspace,), {"asimov": False})]
    assert mock_limit.call_args_list == [
        (("model", "data"), {"tolerance": 0.01, "confidence_level": 0.95})
    ]
    assert mock_vis.call_count == 1
    assert np.allclose(
        mock_vis.call_args[0][0].observed_limit, limit_results.observed_limit
    )
    assert np.allclose(
        mock_vis.call_args[0][0].expected_limit, limit_results.expected_limit
    )
    assert np.allclose(
        mock_vis.call_args[0][0].observed_CLs, limit_results.observed_CLs
    )
    assert np.allclose(
        mock_vis.call_args[0][0].expected_CLs, limit_results.expected_CLs
    )
    assert np.allclose(mock_vis.call_args[0][0].poi_values, limit_results.poi_values)
    assert np.allclose(
        mock_vis.call_args[0][0].confidence_level, limit_results.confidence_level
    )
    assert mock_vis.call_args[1] == {"figure_folder": "figures"}

    # Asimov, tolerance, confidence level, custom folder
    result = runner.invoke(
        cli.limit,
        [
            "--asimov",
            "--tolerance",
            "0.1",
            "--confidence_level",
            "0.90",
            "--figfolder",
            "folder",
            workspace_path,
        ],
    )
    assert result.exit_code == 0
    assert mock_utils.call_args == ((workspace,), {"asimov": True})
    assert mock_limit.call_args == (
        ("model", "data"),
        {"tolerance": 0.1, "confidence_level": 0.90},
    )
    assert mock_vis.call_args[1] == {"figure_folder": "folder"}


@mock.patch("cabinetry.fit.significance", autospec=True)
@mock.patch(
    "cabinetry.model_utils.model_and_data",
    return_value=("model", "data"),
    autospec=True,
)
def test_significance(mock_utils, mock_sig, tmp_path):
    workspace = {"workspace": "mock"}
    workspace_path = str(tmp_path / "workspace.json")

    # need to save workspace to file since click looks for it
    with open(workspace_path, "w") as f:
        f.write('{"workspace": "mock"}')

    runner = CliRunner()

    # default
    result = runner.invoke(cli.significance, [workspace_path])
    assert result.exit_code == 0
    assert mock_utils.call_args_list == [((workspace,), {"asimov": False})]
    assert mock_sig.call_args_list == [(("model", "data"), {})]

    # Asimov
    result = runner.invoke(cli.significance, ["--asimov", workspace_path])
    assert result.exit_code == 0
    assert mock_utils.call_args == ((workspace,), {"asimov": True})
    assert mock_sig.call_args == (("model", "data"), {})


@mock.patch("cabinetry.visualize.data_mc", autospec=True)
@mock.patch(
    "cabinetry.model_utils.prediction", return_value="mock_model_pred", autospec=True
)
@mock.patch(
    "cabinetry.fit.fit",
    return_value=fit.FitResults(
        np.asarray([1.0]), np.asarray([0.1]), ["label"], np.asarray([[1.0]]), 1.0
    ),
    autospec=True,
)
@mock.patch("cabinetry.configuration.validate", autospec=True)
@mock.patch(
    "cabinetry.model_utils.model_and_data",
    return_value=("model", "data"),
    autospec=True,
)
def test_data_mc(
    mock_utils, mock_validate, mock_fit, mock_pred, mock_vis, cli_helpers, tmp_path
):
    workspace = {"workspace": "mock"}
    workspace_path = str(tmp_path / "workspace.json")

    # need to save workspace to file since click looks for it
    with open(workspace_path, "w") as f:
        f.write('{"workspace": "mock"}')

    runner = CliRunner()

    # default
    result = runner.invoke(cli.data_mc, [workspace_path])
    assert result.exit_code == 0
    assert mock_utils.call_args_list == [((workspace,), {})]
    assert mock_validate.call_count == 0
    assert mock_fit.call_count == 0
    assert mock_pred.call_args_list == [(("model",), {"fit_results": None})]
    assert mock_vis.call_args_list == [
        (
            ("mock_model_pred", "data"),
            {
                "config": None,
                "figure_folder": "figures",
                "close_figure": True,
                "save_figure": True,
            },
        )
    ]

    # with config, post-fit, custom figure folder
    config = {"General": {"Measurement": "test_config"}}
    config_path = str(tmp_path / "config.yml")
    cli_helpers.write_config(config_path, config)
    fit_results = fit.FitResults(
        np.asarray([1.0]), np.asarray([0.1]), ["label"], np.asarray([[1.0]]), 1.0
    )

    result = runner.invoke(
        cli.data_mc,
        [workspace_path, "--config", config_path, "--postfit", "--figfolder", "folder"],
    )
    assert result.exit_code == 0
    assert mock_utils.call_args == ((workspace,), {})
    assert mock_validate.call_args_list == [((config,), {})]
    assert mock_fit.call_args_list == [(("model", "data"), {})]
    assert mock_pred.call_args == (("model",), {"fit_results": fit_results})
    assert mock_vis.call_args == (
        ("mock_model_pred", "data"),
        {
            "config": config,
            "figure_folder": "folder",
            "close_figure": True,
            "save_figure": True,
        },
    )


@mock.patch("cabinetry.tabulate.yields", autospec=True)
@mock.patch(
    "cabinetry.model_utils.prediction", return_value="mock_model_pred", autospec=True
)
@mock.patch(
    "cabinetry.fit.fit",
    return_value=fit.FitResults(
        np.asarray([1.0]), np.asarray([0.1]), ["label"], np.asarray([[1.0]]), 1.0
    ),
    autospec=True,
)
@mock.patch(
    "cabinetry.model_utils.model_and_data",
    return_value=("model", "data"),
    autospec=True,
)
def test_yields(mock_utils, mock_fit, mock_pred, mock_tab, tmp_path):
    workspace = {"workspace": "mock"}
    workspace_path = str(tmp_path / "workspace.json")

    # need to save workspace to file since click looks for it
    with open(workspace_path, "w") as f:
        f.write('{"workspace": "mock"}')

    runner = CliRunner()

    # default
    result = runner.invoke(cli.yields, [workspace_path])
    assert result.exit_code == 0
    assert mock_utils.call_args_list == [((workspace,), {})]
    assert mock_fit.call_count == 0
    assert mock_pred.call_args_list == [(("model",), {"fit_results": None})]
    assert mock_tab.call_args_list == [
        (
            ("mock_model_pred", "data"),
            {"table_folder": "tables", "table_format": "simple"},
        )
    ]

    # with post-fit, custom table folder and format
    fit_results = fit.FitResults(
        np.asarray([1.0]), np.asarray([0.1]), ["label"], np.asarray([[1.0]]), 1.0
    )

    result = runner.invoke(
        cli.yields,
        [workspace_path, "--postfit", "--tablefolder", "folder", "--tablefmt", "html"],
    )
    assert result.exit_code == 0
    assert mock_utils.call_args == ((workspace,), {})
    assert mock_fit.call_args_list == [(("model", "data"), {})]
    assert mock_pred.call_args == (("model",), {"fit_results": fit_results})
    assert mock_tab.call_args == (
        ("mock_model_pred", "data"),
        {"table_folder": "folder", "table_format": "html"},
    )


@mock.patch("cabinetry.visualize.modifier_grid", autospec=True)
@mock.patch(
    "cabinetry.model_utils.model_and_data",
    return_value=("model", "data"),
    autospec=True,
)
def test_modifier_grid(mock_utils, mock_vis, cli_helpers, tmp_path):
    workspace = {"workspace": "mock"}
    workspace_path = str(tmp_path / "workspace.json")

    # need to save workspace to file since click looks for it
    with open(workspace_path, "w") as f:
        f.write('{"workspace": "mock"}')

    runner = CliRunner()

    # default
    result = runner.invoke(cli.modifier_grid, [workspace_path])
    assert result.exit_code == 0
    assert mock_utils.call_args_list == [((workspace,), {})]
    assert mock_vis.call_args_list == [
        (
            ("model",),
            {
                "figure_folder": "figures",
                "split_by_sample": False,
                "close_figure": True,
                "save_figure": True,
            },
        )
    ]

    # split by sample, custom figure folder
    result = runner.invoke(
        cli.modifier_grid,
        [workspace_path, "--split_by_sample", "--figfolder", "folder"],
    )
    assert result.exit_code == 0
    assert mock_utils.call_args == ((workspace,), {})
    assert mock_vis.call_args == (
        ("model",),
        {
            "figure_folder": "folder",
            "split_by_sample": True,
            "close_figure": True,
            "save_figure": True,
        },
    )



import logging
import pathlib
from unittest import mock

import boost_histogram as bh
import numpy as np
import pytest

from cabinetry.templates import builder


def test__ntuple_paths(caplog):
    caplog.set_level(logging.DEBUG)

    # only general path, no override
    assert builder._ntuple_paths("path.root", {}, {}, {}, None) == [
        pathlib.Path("path.root")
    ]

    # general path with region and sample templates
    assert builder._ntuple_paths(
        "{RegionPath}/{SamplePath}",
        {"RegionPath": "region"},
        {"SamplePath": "sample.root"},
        {},
        None,
    ) == [pathlib.Path("region/sample.root")]

    # SamplePath with list of two samples
    assert builder._ntuple_paths(
        "{RegionPath}/{SamplePath}",
        {"RegionPath": "region"},
        {"SamplePath": ["sample.root", "new.root"]},
        {},
        None,
    ) == [pathlib.Path("region/sample.root"), pathlib.Path("region/new.root")]

    # systematic with override for RegionPath and SamplePath
    assert builder._ntuple_paths(
        "{RegionPath}/{SamplePath}",
        {"RegionPath": "reg_1"},
        {"SamplePath": "path.root"},
        {
            "Name": "variation",
            "Up": {"SamplePath": ["variation.root", "new.root"], "RegionPath": "reg_2"},
        },
        "Up",
    ) == [pathlib.Path("reg_2/variation.root"), pathlib.Path("reg_2/new.root")]

    # systematic without override
    assert builder._ntuple_paths(
        "{SamplePath}", {}, {"SamplePath": "path.root"}, {"Name": "variation"}, "Up"
    ) == [pathlib.Path("path.root")]

    caplog.clear()

    # warning: no region path in template
    assert builder._ntuple_paths(
        "path.root", {"RegionPath": "region.root"}, {}, {}, None
    ) == [pathlib.Path("path.root")]
    assert "region override specified, but {RegionPath} not found in default path" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # warning: no sample path in template
    assert builder._ntuple_paths(
        "path.root", {}, {"SamplePath": "sample.root"}, {}, None
    ) == [pathlib.Path("path.root")]
    assert "sample override specified, but {SamplePath} not found in default path" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # error: no override for {RegionPath}
    with pytest.raises(ValueError, match="no path setting found for region region"):
        builder._ntuple_paths("{RegionPath}", {"Name": "region"}, {}, {}, None)

    # error: no override for {SamplePath}
    with pytest.raises(ValueError, match="no path setting found for sample sample"):
        builder._ntuple_paths("{SamplePath}", {}, {"Name": "sample"}, {}, None)


def test__variable():
    # no override
    assert builder._variable({"Variable": "jet_pt"}, {}, {}, None) == "jet_pt"

    # systematic with override
    assert (
        builder._variable(
            {"Variable": "jet_pt"},
            {},
            {"Name": "variation", "Up": {"Variable": "jet_pt_up"}},
            "Up",
        )
        == "jet_pt_up"
    )

    # systematic without override
    assert (
        builder._variable({"Variable": "jet_pt"}, {}, {"Name": "variation"}, "Up")
        == "jet_pt"
    )


def test__filter():
    # no override
    assert builder._filter({"Filter": "jet_pt > 0"}, {}, {}, None) == "jet_pt > 0"

    # no filter
    assert builder._filter({}, {}, {}, None) is None

    # systematic with override
    assert (
        builder._filter(
            {"Filter": "jet_pt > 0"},
            {},
            {"Name": "variation", "Up": {"Filter": "jet_pt > 100"}},
            "Up",
        )
        == "jet_pt > 100"
    )

    # systematic without override
    assert (
        builder._filter({"Filter": "jet_pt > 0"}, {}, {"Name": "variation"}, "Up")
        == "jet_pt > 0"
    )

    # sample-specific override
    assert (
        builder._filter({"Filter": "jet_pt > 0"}, {"Filter": "jet_pt > 100"}, {}, None)
        == "jet_pt > 100"
    )

    # sample-specific override, again overridden by systematic
    assert (
        builder._filter(
            {"Filter": "jet_pt > 0"},
            {"Filter": "jet_pt > 100"},
            {"Name": "variation", "Up": {"Filter": "jet_pt > 200"}},
            "Up",
        )
        == "jet_pt > 200"
    )


def test__weight():
    # no override
    assert builder._weight({}, {"Weight": "weight_mc"}, {}, None) == "weight_mc"

    # no weight
    assert builder._weight({}, {}, {}, None) is None

    # systematic with override
    assert (
        builder._weight(
            {},
            {"Weight": "weight_mc"},
            {"Name": "variation", "Up": {"Weight": "weight_modified"}},
            "Up",
        )
        == "weight_modified"
    )

    # systematic without override
    assert (
        builder._weight({}, {"Weight": "weight_mc"}, {"Name": "variation"}, "Up")
        == "weight_mc"
    )


def test__position_in_file():
    # no override
    assert builder._position_in_file({"Tree": "tree_name"}, {}, None) == "tree_name"

    # systematic with override
    assert (
        builder._position_in_file(
            {"Tree": "nominal"}, {"Name": "variation", "Up": {"Tree": "up_tree"}}, "Up"
        )
        == "up_tree"
    )

    # systematic without override
    assert (
        builder._position_in_file({"Tree": "nominal"}, {"Name": "variation"}, "Up")
        == "nominal"
    )


def test__binning():
    np.testing.assert_equal(builder._binning({"Binning": [1, 2, 3]}), [1, 2, 3])
    with pytest.raises(NotImplementedError, match="cannot determine binning"):
        builder._binning({})


def test__Builder():
    builder_instance = builder._Builder(pathlib.Path("path"), "file.root", "uproot")
    assert builder_instance.histogram_folder == pathlib.Path("path")
    assert builder_instance.general_path == "file.root"
    assert builder_instance.method == "uproot"


@mock.patch("cabinetry.templates.utils._name_and_save")
@mock.patch("cabinetry.histo.Histogram", return_value="cabinetry_histogram")
@mock.patch("cabinetry.contrib.histogram_creator.with_uproot", return_value="histogram")
@mock.patch(
    "cabinetry.templates.builder._ntuple_paths",
    return_value=[pathlib.Path("path_to_ntuple")],
)
def test__Builder_create_histogram(
    mock_path, mock_uproot_builder, mock_histo, mock_save
):
    histogram_folder = pathlib.Path("path")
    general_path = "{SamplePath}"
    # the binning [0] is not a proper binning, but simplifies the comparison
    region = {"Name": "test_region", "Variable": "x", "Binning": [0], "Filter": "x>3"}
    sample = {
        "Name": "sample",
        "Tree": "tree",
        "SamplePath": "path_to_sample",
        "Weight": "weight_mc",
    }
    systematic = {}
    template = "Up"

    builder_instance = builder._Builder(histogram_folder, general_path, "uproot")
    builder_instance._create_histogram(region, sample, systematic, template)

    # call to path creation
    assert mock_path.call_args_list == [
        ((general_path, region, sample, systematic, template), {})
    ]

    # verify the backend call happened properly
    assert mock_uproot_builder.call_args_list == [
        (
            ([pathlib.Path("path_to_ntuple")], "tree", "x", [0]),
            {"weight": "weight_mc", "selection_filter": "x>3"},
        )
    ]

    # conversion from bh.Histogram to cabinetry Histogram
    assert mock_histo.call_args_list == [(("histogram",), {})]

    # verify the call for saving wrapped histogram
    assert mock_save.call_args_list == [
        (
            (
                histogram_folder,
                "cabinetry_histogram",
                region,
                sample,
                systematic,
                template,
            ),
            {},
        )
    ]

    # other backends
    builder_unknown = builder._Builder(histogram_folder, "{SamplePath}", "unknown")
    with pytest.raises(NotImplementedError, match="unknown backend unknown"):
        builder_unknown._create_histogram(region, sample, systematic, template)


@mock.patch("cabinetry.templates.utils._name_and_save")
def test__Builder__wrap_custom_template_builder(mock_save):
    histogram = bh.Histogram(bh.axis.Variable([0, 1]))
    region = {"Name": "test_region"}
    sample = {"Name": "sample"}
    systematic = {}
    histogram_folder = pathlib.Path("path")

    def test_func(reg, sam, sys, tem):
        return histogram

    builder_instance = builder._Builder(histogram_folder, "file.root", "uproot")
    wrapped_func = builder_instance._wrap_custom_template_builder(test_func)

    # check the behavior of the wrapped function
    # when called, it should save the returned histogram
    wrapped_func(region, sample, systematic, "Up")

    assert mock_save.call_args_list == [
        ((histogram_folder, histogram, region, sample, systematic, "Up"), {})
    ]

    # wrapped function returns wrong type
    def test_func_wrong_return(reg, sam, sys, tem):
        return None

    wrapped_func_wrong_return = builder_instance._wrap_custom_template_builder(
        test_func_wrong_return
    )
    with pytest.raises(TypeError, match="must return a boost_histogram.Histogram"):
        wrapped_func_wrong_return(region, sample, systematic, "Up")



import logging

import numpy as np
import pytest

import cabinetry
from utils import create_histograms, create_ntuples


@pytest.fixture
def histogram_creator():
    return create_histograms.run


@pytest.fixture
def ntuple_creator():
    return create_ntuples.run


@pytest.mark.no_cover
def test_integration(tmp_path, ntuple_creator, caplog):
    """The purpose of this integration test is to check whether the
    steps run without error and whether the fit result is as expected.
    """
    ntuple_creator(str(tmp_path))

    cabinetry_config = cabinetry.configuration.load("config_example.yml")

    # override config options to point to tmp_path
    cabinetry_config["General"]["HistogramFolder"] = str(tmp_path / "histograms")
    cabinetry_config["General"]["InputPath"] = str(tmp_path / "{SamplePath}")

    caplog.set_level(logging.DEBUG)

    cabinetry.templates.build(cabinetry_config, method="uproot")
    cabinetry.templates.postprocess(cabinetry_config)
    workspace_path = tmp_path / "example_workspace.json"
    ws = cabinetry.workspace.build(cabinetry_config)
    cabinetry.workspace.save(ws, workspace_path)
    ws = cabinetry.workspace.load(workspace_path)
    model, data = cabinetry.model_utils.model_and_data(ws)
    fit_results = cabinetry.fit.fit(
        model, data, minos="Signal_norm", goodness_of_fit=True
    )

    bestfit_expected = [
        -0.32457145,
        -0.58582788,
        1.68953317,
        -0.0880332,
        1.00102289,
        0.98910429,
        1.01970061,
        0.98296235,
    ]
    uncertainty_expected = [
        0.55547829,
        0.62723287,
        0.93885915,
        0.99129285,
        0.04108188,
        0.03792561,
        0.03651562,
        0.04250561,
    ]
    best_twice_nll_expected = 17.194205
    corr_mat_expected = [
        [
            1.0,
            0.94146283,
            -0.93117822,
            0.05351298,
            0.16261406,
            -0.12260987,
            0.0463534,
            -0.04290516,
        ],
        [
            0.94146283,
            1.0,
            -0.8914587,
            -0.1375141,
            0.06767335,
            -0.18879984,
            0.03275504,
            -0.02539973,
        ],
        [
            -0.93117822,
            -0.8914587,
            1.0,
            -0.16827929,
            -0.11133319,
            0.14681713,
            -0.1151505,
            -0.05861778,
        ],
        [
            0.05351298,
            -0.13751410,
            -0.16827929,
            1.0,
            0.01074203,
            -0.02137699,
            0.00170718,
            -0.00075580,
        ],
        [
            0.16261406,
            0.06767335,
            -0.11133319,
            0.01074203,
            1.0,
            0.06280504,
            0.01952484,
            -0.02358537,
        ],
        [
            -0.12260987,
            -0.18879984,
            0.14681713,
            -0.02137699,
            0.06280504,
            1.0,
            0.00906325,
            -0.0049263,
        ],
        [
            0.0463534,
            0.03275504,
            -0.1151505,
            0.00170718,
            0.01952484,
            0.00906325,
            1.0,
            0.06545427,
        ],
        [
            -0.04290516,
            -0.02539973,
            -0.05861778,
            -0.00075580,
            -0.02358537,
            -0.00492630,
            0.06545427,
            1.0,
        ],
    ]
    assert np.allclose(fit_results.bestfit, bestfit_expected)
    assert np.allclose(fit_results.uncertainty, uncertainty_expected, atol=1e-4)
    assert np.allclose(fit_results.best_twice_nll, best_twice_nll_expected)
    assert np.allclose(fit_results.corr_mat, corr_mat_expected, rtol=1e-4, atol=5e-5)
    assert np.allclose(fit_results.goodness_of_fit, 0.24679341)

    # minos result
    assert "Signal_norm                =  1.6895 -0.9580 +0.9052" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # pre- and post-fit yield uncertainties
    prediction_prefit = cabinetry.model_utils.prediction(model)
    assert np.allclose(
        prediction_prefit.total_stdev_model_bins,
        [
            [
                [69.040789, 58.329118, 37.973787, 45.137157],  # background
                [0.0, 0.100772, 1.487773, 1.620867],  # signal
                [69.040789, 58.343328, 38.219599, 45.296964],  # sum over samples
            ]
        ],
    )
    assert np.allclose(
        prediction_prefit.total_stdev_model_channels,
        [[136.368732, 2.851565, 136.791978]],
    )
    _ = cabinetry.visualize.data_mc(prediction_prefit, data, close_figure=True)

    prediction_postfit = cabinetry.model_utils.prediction(
        model, fit_results=fit_results
    )
    assert np.allclose(
        prediction_postfit.total_stdev_model_bins,
        [
            [
                [11.898551, 7.513216, 21.002006, 24.284847],  # background
                [0.0, 1.467646, 22.137293, 22.269200],  # signal
                [11.898551, 7.283171, 7.414715, 7.687966],  # sum over samples
            ]
        ],
        rtol=1e-4,
    )
    assert np.allclose(
        prediction_postfit.total_stdev_model_channels,
        [[41.043814, 45.814417, 20.439575]],
        atol=5e-3,
    )
    _ = cabinetry.visualize.data_mc(prediction_postfit, data, close_figure=True)

    # nuisance parameter ranking
    ranking_results = cabinetry.fit.ranking(
        model, data, fit_results=fit_results, custom_fit=True
    )
    assert np.allclose(
        ranking_results.prefit_up,
        [
            -1.67617185,
            -1.51531885,
            -0.14964691,
            -0.11067885,
            0.14350015,
            -0.11311905,
            -0.05573699,
        ],
    )
    assert np.allclose(
        ranking_results.prefit_down,
        [
            1.41570164,
            1.15001063,
            0.16388633,
            0.10921275,
            -0.15331233,
            0.11492629,
            0.05864962,
        ],
    )
    assert np.allclose(
        ranking_results.postfit_up,
        [
            -1.10235942,
            -0.89697066,
            -0.14842092,
            -0.10774037,
            0.13866107,
            -0.10800505,
            -0.0548506,
        ],
        atol=1e-4,
    )
    assert np.allclose(
        ranking_results.postfit_down,
        [
            0.84366301,
            0.76572728,
            0.16236075,
            0.10604544,
            -0.14383307,
            0.10888311,
            0.05779505,
        ],
        atol=1e-4,
    )

    # parameter scan
    scan_results = cabinetry.fit.scan(
        model,
        data,
        "Signal_norm",
        par_range=(1.18967971, 2.18967971),
        n_steps=3,
        custom_fit=True,
    )
    # lower edge of scan is beyond normalization factor bounds specified in workspace
    assert np.allclose(
        scan_results.delta_nlls, [0.27163227, 0.0, 0.29018620], atol=5e-5
    )

    # upper limit, this calculation is slow
    limit_results = cabinetry.fit.limit(model, data, bracket=(0.5, 3.5), tolerance=0.05)
    assert np.allclose(limit_results.observed_limit, 3.1502, rtol=1e-2)
    assert np.allclose(
        limit_results.expected_limit,
        [1.0054, 1.3975, 1.9689, 2.7174, 3.5426],
        rtol=1e-2,
    )

    # discovery significance
    significance_results = cabinetry.fit.significance(model, data)
    np.allclose(significance_results.observed_p_value, 0.03583662)
    np.allclose(significance_results.observed_significance, 1.80118813)
    np.allclose(significance_results.expected_p_value, 0.14775040)
    np.allclose(significance_results.expected_significance, 1.04613046)


@pytest.mark.no_cover
def test_histogram_reading(tmp_path, histogram_creator):
    histogram_creator(str(tmp_path))

    cabinetry_config = cabinetry.configuration.load("utils/config_histograms.yml")

    # override config options to point to tmp_path
    cabinetry_config["General"]["HistogramFolder"] = str(tmp_path / "histograms")
    cabinetry_config["General"]["InputPath"] = (
        f"{tmp_path / 'histograms.root'}:" + "{RegionPath}/{SamplePath}/{VariationPath}"
    )

    cabinetry.templates.collect(cabinetry_config, method="uproot")
    cabinetry.templates.postprocess(cabinetry_config)
    ws = cabinetry.workspace.build(cabinetry_config)
    model, data = cabinetry.model_utils.model_and_data(ws)
    fit_results = cabinetry.fit.fit(
        model, data, minos="Signal_norm", goodness_of_fit=True
    )

    bestfit_expected = [
        -0.32457145,
        -0.58582788,
        1.68953317,
        -0.0880332,
        1.00102289,
        0.98910429,
        1.01970061,
        0.98296235,
    ]
    best_twice_nll_expected = 17.194205
    assert np.allclose(fit_results.bestfit, bestfit_expected)
    assert np.allclose(fit_results.best_twice_nll, best_twice_nll_expected)



import logging
import pathlib
from unittest import mock

import pytest

from cabinetry.templates import collector


def test__histo_path(caplog):
    caplog.set_level(logging.DEBUG)

    # only general path, no override
    assert collector._histo_path("path.root:h1", "", {}, {}, {}, None) == "path.root:h1"

    # general path with region, sample and nominal variation, template with no impact
    assert (
        collector._histo_path(
            "{RegionPath}.root:{SamplePath}_{VariationPath}",
            "nominal",
            {"RegionPath": "region"},
            {"SamplePath": "sample"},
            {},
            "Up",
        )
        == "region.root:sample_nominal"
    )

    # systematic with override for SamplePath and VariationPath
    assert (
        collector._histo_path(
            "{RegionPath}.root:{SamplePath}_{VariationPath}",
            "nominal",
            {"RegionPath": "reg_1"},
            {"SamplePath": "path"},
            {"Name": "variation", "Up": {"SamplePath": "var", "VariationPath": "up"}},
            "Up",
        )
        == "reg_1.root:var_up"
    )
    caplog.clear()

    # warning: no region path in template
    assert (
        collector._histo_path(
            "f.root:h1", "", {"RegionPath": "region.root"}, {}, {}, None
        )
        == "f.root:h1"
    )
    assert "region override specified, but {RegionPath} not found in default path" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # warning: no sample path in template
    assert (
        collector._histo_path(
            "f.root:h1", "", {}, {"SamplePath": "sample.root"}, {}, None
        )
        == "f.root:h1"
    )
    assert "sample override specified, but {SamplePath} not found in default path" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # warning: no variation path in template
    assert (
        collector._histo_path(
            "f.root:h1", "", {}, {"VariationPath": "sample.root"}, {}, None
        )
        == "f.root:h1"
    )
    assert (
        "variation override specified, but {VariationPath} not found in default path"
        in [rec.message for rec in caplog.records]
    )
    caplog.clear()

    # warning: no colon in path
    assert collector._histo_path("f.root", "", {}, {}, {}, None) == "f.root"
    assert "no colon found in path f.root, may not be able to find histogram" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # error: no override for {RegionPath}
    with pytest.raises(ValueError, match="no path setting found for region region"):
        collector._histo_path("{RegionPath}", "", {"Name": "region"}, {}, {}, None)

    # error: no override for {SamplePath}
    with pytest.raises(ValueError, match="no path setting found for sample sample"):
        collector._histo_path("{SamplePath}", "", {}, {"Name": "sample"}, {}, None)


@mock.patch("cabinetry.templates.utils._name_and_save")
@mock.patch("cabinetry.histo.Histogram", return_value="cabinetry_histogram")
@mock.patch("cabinetry.contrib.histogram_reader.with_uproot", return_value="histogram")
@mock.patch("cabinetry.templates.collector._histo_path", return_value="path_to_hist")
def test__collector(mock_path, mock_uproot_hist, mock_histo, mock_save):
    histogram_folder = pathlib.Path("path")
    general_path = "f.root:{variation_path}"
    variation_path = "nominal"
    processor = collector._collector(
        histogram_folder, general_path, variation_path, "uproot"
    )

    region = {"Name": "region"}
    sample = {"Name": "sample"}
    systematic = {"Name": "systematic"}
    template = "Up"

    # execute processor
    processor(region, sample, systematic, template)

    # call to path creation
    assert mock_path.call_args_list == [
        ((general_path, variation_path, region, sample, systematic, template), {})
    ]

    # call to uproot backend
    assert mock_uproot_hist.call_args_list == [(("path_to_hist",), {})]

    # conversion from bh.Histogram to cabinetry Histogram
    assert mock_histo.call_args_list == [(("histogram",), {})]

    # verify the call for saving wrapped histogram
    assert mock_save.call_args_list == [
        (
            (
                histogram_folder,
                "cabinetry_histogram",
                region,
                sample,
                systematic,
                template,
            ),
            {},
        )
    ]

    # other backends
    processor_unknown = collector._collector(
        histogram_folder, general_path, variation_path, "unknown"
    )
    with pytest.raises(NotImplementedError, match="unknown backend unknown"):
        processor_unknown(region, sample, systematic, None)



"""Implements histogram smoothing algorithms."""

import logging
import statistics
from typing import List, TypeVar, Union

import numpy as np


log = logging.getLogger(__name__)


T = TypeVar("T", List[float], np.ndarray)


def _medians_353(zz: Union[List[float], np.ndarray], nbins: int) -> None:
    """Applies running median smoothing with window sizes 3, 5, 3 to input.

    Args:
        zz (Union[List[float], np.ndarray]): array to smooth
        nbins (int): number of bins in array
    """
    for i_median in range(3):
        yy = zz.copy()  # yy stays constant at each step in the loop
        medianType = 3 if i_median != 1 else 5
        ifirst = 1 if i_median != 1 else 2
        ilast = nbins - 1 if i_median != 1 else nbins - 2
        # do central elements first in (ifirst, ilast) range
        for ii in range(ifirst, ilast):
            zz[ii] = statistics.median(yy[ii - ifirst : ii - ifirst + medianType])

        if i_median == 0:  # first median 3
            # first bin, proceedings use y_1 (=yy[0]), while ROOT uses zz[0]
            zz[0] = statistics.median([3 * zz[1] - 2 * zz[2], yy[0], zz[1]])
            # last bin, proceedings use y_n (=yy[-1]), while ROOT uses zz[-1]
            zz[-1] = statistics.median([zz[-2], yy[-1], 3 * zz[-2] - 2 * zz[-3]])

        if i_median == 1:  # median 5
            zz[1] = statistics.median(yy[0:3])
            # second to last bin
            zz[-2] = statistics.median(yy[-3:])


def smooth_353qh_twice(hist: T) -> T:
    """Runs the 353QH algorithm twice and returns smooth version of the input.

    For documentation see these proceedings https://cds.cern.ch/record/186223/ on page
    292. The algorithm runs twice to avoid over-smoothing peaks and valleys. The
    algorithm is not aware of statistical uncertainties per entry in the array. See also
    https://root.cern.ch/doc/master/classTH1.html#aeb935cae10dbf9cd484bee1b6a549f83 for
    the ROOT implementation.

    Args:
        hist (Union[list, np.ndarray]): array to smooth

    Returns:
        Union[list, np.ndarray]: smooth version of input
    """
    nbins = len(hist)
    if nbins < 3:
        log.warning("at least three points needed for smoothing, no smoothing applied")
        return hist

    if isinstance(hist, np.ndarray):
        # ensure dtype is not int to avoid rounding in smooth histogram
        hist = hist.astype("float")

    zz = np.array(hist, dtype=float, copy=True)

    for i_353QH in range(2):  # run 353QH twice
        # do running median with window sizes 3, 5, 3
        _medians_353(zz, nbins)

        yy = zz.copy()

        # quadratic interpolation for flat segments
        for ii in range(2, nbins - 2):
            if zz[ii - 1] != zz[ii] or zz[ii] != zz[ii + 1]:
                continue  # not flat
            left_larger = zz[ii - 2] - zz[ii]  # two bins left larger by this amount
            right_larger = zz[ii + 2] - zz[ii]  # two bins right larger by this amount
            if left_larger * right_larger <= 0:
                continue  # current position is part of larger scale slope
            jk = 1
            if abs(right_larger) > abs(left_larger):
                jk = -1  # left is more flat than right
            yy[ii] = -0.5 * zz[ii - 2 * jk] + zz[ii] / 0.75 + zz[ii + 2 * jk] / 6
            yy[ii + jk] = 0.5 * (zz[ii + 2 * jk] - zz[ii - 2 * jk]) + zz[ii]

        # running means
        for ii in range(1, nbins - 1):
            zz[ii] = 0.25 * yy[ii - 1] + 0.5 * yy[ii] + 0.25 * yy[ii + 1]
        zz[0] = yy[0]
        zz[-1] = yy[-1]

        if i_353QH == 0:
            # algorithm has been run once
            rr = zz.copy()  # save computed values
            # calculate residuals: (original) - (after 353QH)
            zz = np.asarray([hist[ii] - zz[ii] for ii in range(0, nbins)])
            # zz is now "rough", while rr is "smooth"

        # "twicing": run 353QH again on "rough" zz and add to "smooth" rr

    hist_new = hist.copy()  # histogram after smoothing
    for ii in range(nbins):
        if min(hist) < 0:
            hist_new[ii] = rr[ii] + zz[ii]
        # result is positive if no negative values are in input
        else:
            hist_new[ii] = max(rr[ii] + zz[ii], 0)
    return hist_new
